Package: tweenr
Type: Package
Title: Interpolate Data for Smooth Animations
Version: 2.0.3
Authors@R: 
    c(person(given = "Thomas Lin",
             family = "Pedersen",
             role = c("aut", "cre"),
             email = "thomasp85@gmail.com",
             comment = c(ORCID = "0000-0002-5147-4711")))
Maintainer: Thomas Lin Pedersen <thomasp85@gmail.com>
Description: In order to create smooth animation between states of data,
    tweening is necessary. This package provides a range of functions for
    creating tweened data that can be used as basis for animation. Furthermore 
    it adds a number of vectorized interpolaters for common R data 
    types such as numeric, date and colour.
URL: https://github.com/thomasp85/tweenr
BugReports: https://github.com/thomasp85/tweenr/issues
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 3.2.0)
Imports: farver, magrittr, rlang, vctrs
LinkingTo: cpp11 (>= 0.4.2)
RoxygenNote: 7.2.3
Suggests: testthat, covr
NeedsCompilation: yes
Packaged: 2024-02-26 13:06:33 UTC; thomas
Author: Thomas Lin Pedersen [aut, cre]
    (<https://orcid.org/0000-0002-5147-4711>)
Repository: CRAN
Date/Publication: 2024-02-26 13:50:03 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-27 14:16:21 UTC; unix
Archs: tweenr.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(convert_generator,along_generator)
S3method(convert_generator,component_generator)
S3method(convert_generator,default)
S3method(convert_generator,keyframe_generator)
S3method(get_frame,along_generator)
S3method(get_frame,component_generator)
S3method(get_frame,keyframe_generator)
S3method(get_raw_frames,along_generator)
S3method(get_raw_frames,component_generator)
S3method(get_raw_frames,keyframe_generator)
S3method(vec_tween_class,Date)
S3method(vec_tween_class,POSIXt)
S3method(vec_tween_class,character)
S3method(vec_tween_class,default)
S3method(vec_tween_class,factor)
S3method(vec_tween_class,list)
S3method(vec_tween_class,logical)
S3method(vec_tween_class,numeric)
export("%>%")
export("col_types<-")
export("ease_type<-")
export("frame_times<-")
export("gen_data<-")
export("generator_settings<-")
export("keyframes<-")
export(.complete_states)
export(.get_first_frame)
export(.get_last_frame)
export(.has_frames)
export(.max_id)
export(.with_later_frames)
export(.with_prior_frames)
export(PHASE_LEVELS)
export(add_keyframe)
export(add_pause)
export(close_state)
export(col_classes)
export(col_types)
export(data_frame_to_gen)
export(display_ease)
export(ease_type)
export(frame_times)
export(gen_along)
export(gen_at)
export(gen_components)
export(gen_data)
export(gen_events)
export(gen_keyframe)
export(gen_to_data_frame)
export(generator_settings)
export(get_frame)
export(get_raw_frames)
export(interpolate_custom_at)
export(is_generator)
export(keep_state)
export(keyframes)
export(open_state)
export(prepare_keyframes)
export(tween)
export(tween_along)
export(tween_appear)
export(tween_at)
export(tween_at_t)
export(tween_color)
export(tween_color_t)
export(tween_colour)
export(tween_colour_t)
export(tween_components)
export(tween_constant)
export(tween_constant_t)
export(tween_date)
export(tween_date_t)
export(tween_datetime)
export(tween_datetime_t)
export(tween_elements)
export(tween_events)
export(tween_fill)
export(tween_numeric)
export(tween_numeric_t)
export(tween_state)
export(tween_states)
export(tween_t)
export(vec_tween_class)
importFrom(farver,decode_colour)
importFrom(farver,encode_colour)
importFrom(graphics,plot)
importFrom(magrittr,"%>%")
importFrom(rlang,"%||%")
importFrom(rlang,as_function)
importFrom(rlang,enquo)
importFrom(rlang,eval_tidy)
importFrom(rlang,is_integerish)
importFrom(rlang,quo_is_missing)
importFrom(rlang,quo_is_null)
importFrom(vctrs,vec_cbind)
importFrom(vctrs,vec_rbind)
useDynLib(tweenr, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# tweenr <img src="man/figures/logo.png" align="right" />

<!-- badges: start -->

[![R-CMD-check](https://github.com/thomasp85/tweenr/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/thomasp85/tweenr/actions/workflows/R-CMD-check.yaml)
[![CRAN_Release_Badge](http://www.r-pkg.org/badges/version-ago/tweenr)](https://CRAN.R-project.org/package=tweenr)
[![CRAN_Download_Badge](http://cranlogs.r-pkg.org/badges/tweenr)](https://CRAN.R-project.org/package=tweenr)
<!-- badges: end -->

## What is this?

`tweenr` is a package for interpolating data, mainly for animations. It
provides a range of functions that take data of different forms and
calculate intermediary values. It supports all atomic vector types along
with `factor`, `Date`, `POSIXct`, characters representing colours, and
`list`. `tweenr` is used extensibly by
[`gganimate`](https://github.com/thomasp85/gganimate) to create smooth
animations, but can also be used by itself to prepare data for animation
in another framework.

## How do I get it?

`tweenr` is available on CRAN and can be installed with
`install.packages('tweenr')`. In order to get the development version
you can install it from github with `devtools`

``` r
#install.packages('devtools')
devtools::install_github('thomasp85/tweenr')
```

## An example

Following is an example of using the pipeable `tween_state()` function
with our belowed iris data:

``` r
library(tweenr)
library(ggplot2)

# Prepare the data with some extra columns
iris$col <- c('firebrick', 'forestgreen', 'steelblue')[as.integer(iris$Species)]
iris$size <- 4
iris$alpha <- 1
iris <- split(iris, iris$Species)

# Here comes tweenr
iris_tween <- iris$setosa %>% 
  tween_state(iris$versicolor, ease = 'cubic-in-out', nframes = 30) %>% 
  keep_state(10) %>% 
  tween_state(iris$virginica, ease = 'elastic-out', nframes = 30) %>% 
  keep_state(10) %>% 
  tween_state(iris$setosa, ease = 'quadratic-in', nframes = 30) %>% 
  keep_state(10)

# Animate it to show the effect
p_base <- ggplot() + 
  geom_point(aes(x = Petal.Length, y = Petal.Width, alpha = alpha, colour = col, 
                 size = size)) + 
  scale_colour_identity() +
  scale_alpha_identity() + 
  scale_size_identity() + 
  coord_cartesian(xlim = range(iris_tween$Petal.Length), 
                  ylim = range(iris_tween$Petal.Width))
iris_tween <- split(iris_tween, iris_tween$.frame)
for (d in iris_tween) {
  p <- p_base %+% d
  plot(p)
}
```

![](man/figures/README-unnamed-chunk-3.gif)

## Other functions

Besides the `tween_state()`/`keep_state()` combo showcased above, there
are a slew of other functions meant for data in different formats

**`tween_components`** takes a single data.frame, a vector of ids
identifying recurrent elements, and a vector of timepoints for each row
and interpolate each element between its specified time points.

**`tween_events`** takes a single data.frame where each row encodes a
single unique event, along with a start, and end time and expands the
data across a given number of frames.

**`tween_along`** takes a single data.frame along with an id and
timepoint vector and calculate evenly spaced intermediary values with
the possibility of keeping old values at each frame.

**`tween_at`** takes two data.frames or vectors along with a numeric
vector giving the interpolation point between the two data.frames to
calculate.

**`tween_fill`** fills missing values in a vector or data.frame by
interpolating between previous and next non-missing elements

## Easing

In order to get smooth transitions you’d often want a non-linear
interpolation. This can be achieved by using an easing function to
translate the equidistant interpolation points into new ones. `tweenr`
has support for a wide range of different easing functions, all of which
can be previewed using `display_ease()` as here where the popular
*cubic-in-out* is shown:

``` r
tweenr::display_ease('cubic-in-out')
```

![](man/figures/README-unnamed-chunk-4-1.png)<!-- -->

## Spatial interpolations

The purpose of `tweenr` is to interpolate values independently. If paths
and polygons needs to be transitioned the
[`transformr`](https://github.com/thomasp85/transformr) package should
be used as it expands tweenr into the spatial realm

#' @rdname tween
#'
#' @export
tween_constant <- function(data, n, ease = 'linear') {
  data <- as.list(data)
  data <- lapply(data, as.character)
  prepData <- prepareTween(data, n, ease)
  tweendata <- do.call(interpolate_character_state, prepData)
  unname(split(tweendata,
               rep(seq_along(data[[1]]), length.out = length(tweendata))))
}

#' @rdname tween
#'
#' @export
tween_constant_t <- function(data, n, ease = 'linear') {
  if (!is.list(data)) {
    data <- list(data)
  }
  data <- lapply(data, as.character)
  prepData <- prepareTweenTranspose(data, n, ease)
  tweendata <- do.call(interpolate_character_state, prepData)
  unname(split(tweendata,
               rep(seq_along(data), rep(n, length.out = length(data)))))
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

numeric_state_interpolator <- function(data, states) {
    .Call('_tweenr_numeric_state_interpolator', PACKAGE = 'tweenr', data, states)
}

colour_state_interpolator <- function(data, states) {
    .Call('_tweenr_colour_state_interpolator', PACKAGE = 'tweenr', data, states)
}

constant_state_interpolator <- function(data, states) {
    .Call('_tweenr_constant_state_interpolator', PACKAGE = 'tweenr', data, states)
}

list_state_interpolator <- function(data, states) {
    .Call('_tweenr_list_state_interpolator', PACKAGE = 'tweenr', data, states)
}

numlist_state_interpolator <- function(data, states) {
    .Call('_tweenr_numlist_state_interpolator', PACKAGE = 'tweenr', data, states)
}

phase_state_interpolator <- function(data, states) {
    .Call('_tweenr_phase_state_interpolator', PACKAGE = 'tweenr', data, states)
}

numeric_element_interpolator <- function(data, group, frame, ease) {
    .Call('_tweenr_numeric_element_interpolator', PACKAGE = 'tweenr', data, group, frame, ease)
}

colour_element_interpolator <- function(data, group, frame, ease) {
    .Call('_tweenr_colour_element_interpolator', PACKAGE = 'tweenr', data, group, frame, ease)
}

constant_element_interpolator <- function(data, group, frame, ease) {
    .Call('_tweenr_constant_element_interpolator', PACKAGE = 'tweenr', data, group, frame, ease)
}

list_element_interpolator <- function(data, group, frame, ease) {
    .Call('_tweenr_list_element_interpolator', PACKAGE = 'tweenr', data, group, frame, ease)
}

numlist_element_interpolator <- function(data, group, frame, ease) {
    .Call('_tweenr_numlist_element_interpolator', PACKAGE = 'tweenr', data, group, frame, ease)
}

phase_element_interpolator <- function(data, group, frame, ease) {
    .Call('_tweenr_phase_element_interpolator', PACKAGE = 'tweenr', data, group, frame, ease)
}

numeric_along_interpolator <- function(data, group, time, history, keep_last, nframes, ease) {
    .Call('_tweenr_numeric_along_interpolator', PACKAGE = 'tweenr', data, group, time, history, keep_last, nframes, ease)
}

colour_along_interpolator <- function(data, group, time, history, keep_last, nframes, ease) {
    .Call('_tweenr_colour_along_interpolator', PACKAGE = 'tweenr', data, group, time, history, keep_last, nframes, ease)
}

constant_along_interpolator <- function(data, group, time, history, keep_last, nframes, ease) {
    .Call('_tweenr_constant_along_interpolator', PACKAGE = 'tweenr', data, group, time, history, keep_last, nframes, ease)
}

list_along_interpolator <- function(data, group, time, history, keep_last, nframes, ease) {
    .Call('_tweenr_list_along_interpolator', PACKAGE = 'tweenr', data, group, time, history, keep_last, nframes, ease)
}

numlist_along_interpolator <- function(data, group, time, history, keep_last, nframes, ease) {
    .Call('_tweenr_numlist_along_interpolator', PACKAGE = 'tweenr', data, group, time, history, keep_last, nframes, ease)
}

phase_along_interpolator <- function(group, time, history, keep_last, nframes) {
    .Call('_tweenr_phase_along_interpolator', PACKAGE = 'tweenr', group, time, history, keep_last, nframes)
}

numeric_at_interpolator <- function(from, to, at, ease) {
    .Call('_tweenr_numeric_at_interpolator', PACKAGE = 'tweenr', from, to, at, ease)
}

colour_at_interpolator <- function(from, to, at, ease) {
    .Call('_tweenr_colour_at_interpolator', PACKAGE = 'tweenr', from, to, at, ease)
}

constant_at_interpolator <- function(from, to, at, ease) {
    .Call('_tweenr_constant_at_interpolator', PACKAGE = 'tweenr', from, to, at, ease)
}

list_at_interpolator <- function(from, to, at, ease) {
    .Call('_tweenr_list_at_interpolator', PACKAGE = 'tweenr', from, to, at, ease)
}

numlist_at_interpolator <- function(from, to, at, ease) {
    .Call('_tweenr_numlist_at_interpolator', PACKAGE = 'tweenr', from, to, at, ease)
}

numeric_fill_interpolator <- function(data, ease) {
    .Call('_tweenr_numeric_fill_interpolator', PACKAGE = 'tweenr', data, ease)
}

colour_fill_interpolator <- function(data, ease) {
    .Call('_tweenr_colour_fill_interpolator', PACKAGE = 'tweenr', data, ease)
}

constant_fill_interpolator <- function(data, ease) {
    .Call('_tweenr_constant_fill_interpolator', PACKAGE = 'tweenr', data, ease)
}

list_fill_interpolator <- function(data, ease) {
    .Call('_tweenr_list_fill_interpolator', PACKAGE = 'tweenr', data, ease)
}

numlist_fill_interpolator <- function(data, ease) {
    .Call('_tweenr_numlist_fill_interpolator', PACKAGE = 'tweenr', data, ease)
}

#' Transition in and out of events
#'
#' This tweening function is a more powerful version of [tween_appear()], with
#' support for newer features such as enter/exits and tween phase
#' identification. The tweener treats each row in the data as unique events in
#' time, and creates frames with the correct events present at any given time.
#'
#' @param start,end The start (and potential end) of the event encoded in the
#' row, as unquoted expressions. Will be evaluated in the context of `.data` so
#' can refer to columns in it. If `end = NULL` the event will be without extend
#' and only visible in a single frame, unless `enter` and/or `exit` is given.
#'
#' @inheritParams tween_components
#'
#' @return A data.frame with the same columns as `.data` along with `.id` giving
#' the component id, `.phase` giving the state of each component in each frame,
#' and `.frame` giving the frame membership of each row.
#'
#' @family data.frame tween
#'
#' @importFrom rlang enquo quo_is_missing eval_tidy
#' @export
#'
#' @examples
#' d <- data.frame(
#'   x = runif(20),
#'   y = runif(20),
#'   time = runif(20),
#'   duration = runif(20, max = 0.1)
#' )
#' from_left <- function(x) {
#'   x$x <- -0.5
#'   x
#' }
#' to_right <- function(x) {
#'   x$x <- 1.5
#'   x
#' }
#'
#' tween_events(d, 'cubic-in-out', 50, start = time, end = time + duration,
#'              enter = from_left, exit = to_right, enter_length = 0.1,
#'              exit_length = 0.05)
#'
tween_events <- function(.data, ease, nframes, start, end = NULL, range = NULL, enter = NULL, exit = NULL, enter_length = 0, exit_length = 0) {
  start <- enquo(start)
  if (quo_is_missing(start)) stop('start must be provided', call. = FALSE)
  start <- eval_tidy(start, .data)
  end <- enquo(end)
  end <- eval_tidy(end, .data)
  enter_length <- enquo(enter_length)
  enter_length <- eval_tidy(enter_length, .data)
  exit_length <- enquo(exit_length)
  exit_length <- eval_tidy(exit_length, .data)

  if (is.null(enter_length)) enter_length <- 0
  if (is.null(exit_length)) exit_length <- 0
  .data <- .complete_events(.data, start, end, enter, exit, enter_length, exit_length)

  .tween_individuals(.data, ease, nframes, range)
}

#' @importFrom vctrs vec_rbind
#' @importFrom rlang as_function
.complete_events <- function(data, start, end, enter, exit, enter_length, exit_length) {
  data$.id <- seq_len(nrow(data))
  data$.phase <- rep("raw", nrow(data))
  start <- rep(start, length.out = nrow(data))
  if (is.null(end)) {
    event_end <- data[0, , drop = FALSE]
    end <- start[0]
  } else {
    event_end <- data
    end <- rep(end, length.out = nrow(data))
    data$.phase <- 'static'
  }
  if (is.null(enter)) {
    enter_data <- data[0, , drop = FALSE]
    enter_time <- start[0]
  } else {
    enter_data <- as_function(enter)(data)
    enter_data$.phase <- 'enter'
    enter_time <- start - enter_length
  }
  if (is.null(exit)) {
    exit_data <- data[0, , drop = FALSE]
    exit_time <- start[0]
  } else {
    exit_data <- as_function(exit)(data)
    exit_data$.phase <- 'exit'
    exit_time <- (if (length(end) == 0) start else end) + exit_length
  }
  data <- vec_rbind(enter_data, data, event_end, exit_data)
  time <- c(enter_time, start, end, exit_time)
  data$.time <- time
  data
}

context("events")

df <- data.frame(
  x = c(1, 5, 7, 10),
  y = c(4, 3, 7, -1),
  col = c('black', 'red', 'green', 'blue'),
  type = letters[1:4],
  stringsAsFactors = FALSE
)

test_that("tween_events works", {
  tween <- tween_events(df, 'linear', 20, x, x + 2)
  expect_equal(nrow(tween), 17)
  expect_equal(max(tween$.frame), 20)
  expect_true(all(tween$.phase[c(4, 8, 13, 17)] == 'raw'))
  expect_true(all(tween$.phase[-c(4, 8, 13, 17)] == 'static'))


  tween <- tween_events(df, 'linear', 20, x, enter = function(df) {
    df$x <- 0
    df$col <- 'red'
    df
  }, enter_length = 3)


  expect_equal(nrow(tween), 23)
  expect_equal(max(tween$.frame), 20)
  expect_true(all(tween$.phase[c(6, 13, 17, 23)] == 'raw'))
  expect_true(all(tween$.phase[-c(6, 13, 17, 23)] == 'enter'))
  expect_equal(tween$x[2], 0.2)
  expect_equal(tween$col[3], '#931B0B')
})

test_that("weird input gets handled", {
  expect_error(tween_events(df, 'linear', 20))
  tween <- tween_events(df, 'linear', 0, x)
  expect_equal(nrow(tween), 0)
  tween <- tween_events(df[integer(), ], 'linear', 10, x)
  expect_equal(nrow(tween), 0)
})

#' Tween a list of data.frames representing states
#'
#' This function is intended to create smooth transitions between states of
#' data. States are defined as full data.frames or data.frames containing only
#' the columns with change. Each state can have a defined period of pause, the
#' transition length between each states can be defined as well as the easing
#' function.
#'
#' @param data A list of data.frames. Each data.frame must contain the same
#' number of rows, but only the first data.frame needs to contain all columns.
#' Subsequent data.frames need only contain the columns that shows change.
#'
#' @param tweenlength The lengths of the transitions between each state.
#'
#' @param statelength The length of the pause at each state.
#'
#' @param ease The easing functions to use for the transitions. See details.
#'
#' @param nframes The number of frames to generate. The actual number of frames
#' might end up being higher depending on the regularity of `tweenlength`
#' and `statelength`.
#'
#' @return A data.frame with the same columns as the first data.frame in
#' `data`, but replicated `nframes` times. An additional column called
#' `.frame` will be added giving the frame number.
#'
#' @family data.frame tween
#'
#' @importFrom vctrs vec_cbind
#'
#' @examples
#' data1 <- data.frame(
#'   x = 1:20,
#'   y = 0,
#'   colour = 'forestgreen',
#'   stringsAsFactors = FALSE
#' )
#' data2 <- data1
#' data2$x <- 20:1
#' data2$y <- 1
#'
#' data <- tween_states(list(data1, data2), 3, 1, 'cubic-in-out', 100)
#'
#' @export
#'
tween_states <- function(data, tweenlength, statelength, ease, nframes) {
  if (!(is.list(data) && all(sapply(data, is.data.frame)))) {
    stop('data must be a list of data.frames')
  }
  if (length(data) == 1) {
    stop('data must contain multiple states')
  }
  if (length(unique(sapply(data, nrow))) != 1) {
    stop('All elements in data must have the same number of rows')
  }
  data <- lapply(data, function(d) {
    d$.phase <- 'raw'
    d
  })
  origNames <- names(data[[1]])
  if (!is.list(ease)) ease <- as.list(ease)
  allNames <- unlist(lapply(data, names))
  if (!all(allNames %in% origNames)) {
    stop('All columns must be specified in the original data.frame')
  }
  nstates <- length(data)
  tweenlength <- rep(tweenlength, nstates)[seq_len(nstates - 1)]
  statelength <- rep(statelength, nstates)[seq_len(nstates)]
  ease <- rep(ease, nstates)[seq_len(nstates - 1)]
  pauseIndex <- which(rep(c(TRUE, FALSE), length.out = 2*nstates - 1))
  tweenIndex <- which(rep(c(FALSE, TRUE), length.out = 2*nstates - 1))
  statesOrder <- order(c(pauseIndex, tweenIndex))
  states <- data.frame(
    length = c(statelength, tweenlength)[statesOrder],
    nframes = NA_integer_,
    state = NA_integer_,
    stringsAsFactors = FALSE
  )
  states$state <- rep(seq_len(nstates) - 1L, each = 2, length.out = nrow(states))
  states$ease <- lapply(c(rep(list('constant'), nstates), ease)[statesOrder], function(e) {
    structure(rep(e, length.out = length(origNames)), names = origNames)
  })
  fullLength <- sum(states$length)
  framelength <- fullLength/nframes
  states$nframes <- as.integer(round(states$length / framelength))
  nframes <- sum(states$nframes)
  framelength <- fullLength/nframes
  data <- Reduce(function(l, r) {
    extraCols <- !names(l[[length(l)]]) %in% names(r);
    append(l, list(vec_cbind(r, l[[length(l)]][, extraCols])))
  }, data[-1], data[1])
  colClasses <- col_classes(data[[1]])
  tweendata <- lapply(names(data[[1]]),  function(name) {
    d <- lapply(data, `[[`, i = name)
    d_states <- states
    d_states$ease <- vapply(d_states$ease, `[`, character(1), i = name)
    switch(
      colClasses[name],
      numeric = interpolate_numeric_state(d, d_states),
      logical = interpolate_logical_state(d, d_states),
      factor = interpolate_factor_state(d, d_states),
      character = interpolate_character_state(d, d_states),
      colour = interpolate_colour_state(d, d_states),
      date = interpolate_date_state(d, d_states),
      datetime = interpolate_datetime_state(d, d_states),
      constant = interpolate_constant_state(d, d_states),
      numlist = interpolate_numlist_state(d, d_states),
      list = interpolate_list_state(d, d_states),
      phase = get_phase_state(d, d_states)
    )
  })
  tweendata <- structure(tweendata, names = names(data[[1]]), row.names = .set_row_names(length(tweendata[[1]])), class = 'data.frame')
  tweendata$.id <- rep(seq_len(nrow(data[[1]])), each = nframes)
  tweendata$.frame <- rep(seq_len(nframes), each = nrow(data[[1]]))
  attr(tweendata, 'framelength') <- framelength
  tweendata
}

#' Generator for keyframe based tweening
#'
#' This is a generator version of [tween_state()] and its utility functions. It
#' returns a generator that can be used with [get_frame()] and
#' [get_raw_frames()] to extract frames for a specific time point scaled between
#' 0 and 1.
#'
#' @param keyframe A data frame to use as a keyframe state
#' @inheritParams tween_state
#' @param pause The length of the pause at the current keyframe
#' @param length The length of the transition
#'
#' @return A `keyframe_generator` object
#'
#' @family Other generators
#'
#' @importFrom vctrs vec_cbind
#' @export
#'
#' @examples
#' df1 <- data.frame(
#'   country = c('Denmark', 'Sweden', 'Norway'),
#'   population = c(5e6, 10e6, 3.5e6)
#' )
#' df2 <- data.frame(
#'   country = c('Denmark', 'Sweden', 'Norway', 'Finland'),
#'   population = c(6e6, 10.5e6, 4e6, 3e6)
#' )
#' df3 <- data.frame(
#'   country = c('Denmark', 'Norway'),
#'   population = c(10e6, 6e6)
#' )
#' to_zero <- function(x) {
#'   x$population <- 0
#'   x
#' }
#' gen <- gen_keyframe(df1, 10) %>%
#'   add_keyframe(df2, 'cubic-in-out', 35, id = country, enter = to_zero) %>%
#'   add_pause(10) %>%
#'   add_keyframe(df3, 'cubic-in-out', 35, id = country, enter = to_zero,
#'                exit = to_zero) %>%
#'   add_pause(10)
#'
#' get_frame(gen, 0.25)
gen_keyframe <- function(keyframe = NULL, pause = 0) {
  if (is.null(keyframe)) {
    keyframe <- data.frame()
  }
  if (!is.data.frame(keyframe)) {
    stop("`start` must be `NULL` or a data frame", call. = FALSE)
  }
  keyframe$.phase <- rep_len(factor("raw", levels = PHASE_LEVELS), nrow(keyframe))
  class(keyframe) <- c(c("keyframe_generator", "frame_generator"), class(keyframe))
  generator_settings(keyframe) <- list(
    keyframes = list(list(keyframe, keyframe)),
    frame_time = c(0, pause),
    ease_type = list(),
    col_types = col_classes(keyframe)
  )
  keyframe
}

#' @rdname gen_keyframe
#' @export
add_pause <- function(.data, pause = 0) {
  cur_frame_time <- frame_times(.data)
  if (has_last_pause(.data)) {
    last(cur_frame_time) <- last(cur_frame_time) + pause
  } else {
    cur_frame_time <- c(cur_frame_time, pause + last(cur_frame_time))
  }
  frame_times(.data) <- cur_frame_time
  .data
}
#' @rdname gen_keyframe
#' @importFrom vctrs vec_cbind vec_rbind
#' @importFrom rlang enquo
#' @export
add_keyframe <- function(.data, keyframe, ease, length, id = NULL, enter = NULL, exit = NULL) {
  id <- enquo(id)
  .data <- prepare_keyframes(.data, keyframe)
  from <- last(keyframes(.data))[[2]]
  keyframe$.phase = rep_len(factor("raw", levels = PHASE_LEVELS), nrow(keyframe))
  full <- vec_rbind(from, keyframe)
  full <- .complete_states(
    full[seq_len(nrow(from)), ],
    full[-seq_len(nrow(from)), ],
    id,
    enter,
    exit,
    0
  )
  last(keyframes(.data))[[3]] <- full$from
  keyframes(.data) <- c(keyframes(.data), list(list(full$to, keyframe)))
  frame_times(.data) <- c(frame_times(.data), length + last(frame_times(.data)))
  ease_type(.data) <- c(ease_type(.data), list(ease))
  .data <- structure(
    vec_rbind(.data, keyframe),
    class = class(.data),
    generator_settings = generator_settings(.data)
  )
  col_types(.data) <- col_classes(.data)
  .data
}

#' @export
get_frame.keyframe_generator <- function(generator, at, ...) {
  # clamp between 0 and 1
  at <- min(max(at, 0), 1)
  ft <- frame_times(generator)

  # normalise to generator time
  at <- at * last(ft)

  # Find start stage
  stage <- max(first(which(at <= ft)) - 1, 1)
  if (at == 0) stage <- stage + 1
  key <- (stage + 1) %/% 2

  # if in pause, return the unaltered data (with "static" phase)
  if (stage %% 2 == 1 && !at %in% ft) {
    frame <- keyframes(generator)[[key]][[2]]
    frame$.phase <- factor("static", levels = PHASE_LEVELS)
    return(frame)
  }

  ease <- ease_type(generator)[[key]]
  pos <- (at - ft[stage]) / (ft[stage + 1] - ft[stage])
  frame_at(keyframes(generator)[[key]][[3]], keyframes(generator)[[key + 1]][[1]], pos, ease, col_types(generator))
}
#' @export
#' @importFrom vctrs vec_rbind
get_raw_frames.keyframe_generator <- function(generator, at, before = 0, after = 0, ...) {
  # clamp between 0 and 1
  before <- min(max(at - before, 0), 1)
  after <- min(max(at + after, 0), 1)
  at <- min(max(at, 0), 1)
  ft <- frame_times(generator)

  # normalise to generator time
  at <- at * last(ft)
  before <- before * last(ft)
  after <- after * last(ft)

  # Find before and after
  before <- which(ft >= before & ft < at)
  before <- unique((before + 1) %/% 2)
  after <- which(ft > at & ft <= after)
  after <- unique((after + 1) %/% 2)

  list(
    before = vec_rbind(lapply(keyframes(generator)[before], `[[`, 2)),
    after = vec_rbind(lapply(keyframes(generator)[after], `[[`, 2))
  )
}

#' @export
convert_generator.keyframe_generator <- function(x) {
  data <- keyframes(x)
  outer_id <- rep(seq_along(data), lengths(data))
  data <- unlist(data, recursive = FALSE)
  inner_id <- rep(seq_along(data), vapply(data, nrow, integer(1)))
  data <- vec_rbind(data)
  settings <- list(
    attributes = attributes(x),
    data = data.frame(x),
    inner_id = inner_id,
    outer_id = outer_id,
    converter = function(x, settings) {
      data <- settings$data
      attributes(data) <- settings$attributes
      keyframes(data) <- split(split(data.frame(x), settings$inner_id), settings$outer_id)
      data
    }
  )
  list(data = data, settings = settings)
}

# Utils -------------------------------------------------------------------

#' Prepare keyframes for generator
#'
#' @param .data A keyframe generator
#' @param keyframe A keyframe to add
#'
#' @return A valid keyframe generator
#'
#' @export
#' @keywords internal
#'
prepare_keyframes <- function(.data, keyframe) {
  if (!is.data.frame(keyframe)) {
    stop("`keyframe` must be a data frame", call. = FALSE)
  }
  if (!is_keyframe_generator(.data)) {
    stop("`.data` must be a keyframe generator", call. = FALSE)
  }
  if (is_keyframe_placeholder(.data)) {
    new_start <- generator_settings(keyframe[0, ])
    frame_times(new_start) <- frame_times(.data)
    .data <- new_start
  }
  if (!has_last_pause(.data)) {
    .data <- add_pause(.data, 0)
  }
  .data
}

is_keyframe_generator <- function(x) inherits(x, "keyframe_generator")
is_keyframe_placeholder <- function(x) isTRUE(names(x) == ".phase") && nrow(x) == 0

#' @rdname gen_internal
#' @export
keyframes <- function(x) generator_settings(x)$keyframes

#' @rdname gen_internal
#' @export
`keyframes<-` <- function(x, value) {
  generator_settings(x)$keyframes <- value
  x
}

#' @rdname gen_internal
#' @export
frame_times <- function(x) generator_settings(x)$frame_time

#' @rdname gen_internal
#' @export
`frame_times<-` <- function(x, value) {
  generator_settings(x)$frame_time <- value
  x
}

has_last_pause <- function(x) (length(frame_times(x)) %% 2) == 0

get_phase_at <- function(from, to) {
  factor(
    ifelse(from == "enter", "enter", ifelse(to == "exit", "exit", "transition")),
    levels = PHASE_LEVELS
  )
}

frame_at <- function(from, to, at, ease, classes) {
  if (at == 0) return(from)
  if (at == 1) return(to)
  at <- rep_len(at, nrow(from))
  ease <- rep_len(ease, ncol(from))
  data <- lapply(seq_along(classes), function(i) {
    switch(
      classes[i],
      numeric = interpolate_numeric_at(from[[i]], to[[i]], at, ease[i]),
      logical = interpolate_logical_at(from[[i]], to[[i]], at, ease[i]),
      factor = interpolate_factor_at(from[[i]], to[[i]], at, ease[i]),
      character = interpolate_character_at(from[[i]], to[[i]], at, ease[i]),
      colour = interpolate_colour_at(from[[i]], to[[i]], at, ease[i]),
      date = interpolate_date_at(from[[i]], to[[i]], at, ease[i]),
      datetime = interpolate_datetime_at(from[[i]], to[[i]], at, ease[i]),
      constant = interpolate_constant_at(from[[i]], to[[i]], at, ease[i]),
      numlist = interpolate_numlist_at(from[[i]], to[[i]], at, ease[i]),
      list = interpolate_list_at(from[[i]], to[[i]], at, ease[i]),
      phase = get_phase_at(from[[i]], to[[i]]),
      interpolate_custom_at(from[[i]], to[[i]], at, ease[i])
    )
  })
  structure(data, names = names(from), row.names = .set_row_names(length(data[[1]])), class = 'data.frame')
}

#' Fallback for keyframe vector support
#'
#' @param from,to vectors to interpolate between
#' @param at value between 0 and 1 defining the point
#' @param ease the easing function to use
#'
#' @export
#' @keywords internal
interpolate_custom_at <- function(from, to, at, ease) {
  UseMethod('interpolate_custom_at')
}

interpolate_numeric_element <- function(data, group, frame, ease) {
  numeric_element_interpolator(as.numeric(data), as.integer(group), as.integer(frame), as.character(ease))
}

interpolate_logical_element <- function(data, group, frame, ease) {
  res <- interpolate_numeric_element(data, group, frame, ease)
  res[['data']] <- as.logical(round(res[['data']]))
  res
}

#' @importFrom farver decode_colour encode_colour
interpolate_colour_element <- function(data, group, frame, ease) {
  data <- decode_colour(data, alpha = TRUE, to = 'lab')
  col <- colour_element_interpolator(data, as.integer(group), as.integer(frame), as.character(ease))
  data.frame(
    data = encode_colour(col[, 1:3, drop = FALSE], alpha = col[,4], from = 'lab'),
    group = col$group,
    frame = col$frame,
    stringsAsFactors = FALSE
  )
}

interpolate_constant_element <- function(data, group, frame, ease) {
  constant_element_interpolator(as.character(data), as.integer(group), as.integer(frame), as.character(ease))
}

interpolate_character_element <- interpolate_constant_element

interpolate_date_element <- function(data, group, frame, ease) {
  res <- interpolate_numeric_element(data, group, frame, ease)
  res[['data']] <- as.Date(res[['data']], origin = BASEDATE)
  res
}

interpolate_datetime_element <- function(data, group, frame, ease) {
  if (inherits(data, 'POSIXlt')) {
    warning("POSIXlt converted to POSIXct")
    data <- as.POSIXct(data)
  }
  tz <- attr(data, 'tzone')
  res <- interpolate_numeric_element(data, group, frame, ease)
  res[['data']] <-  as.POSIXct(res[['data']], origin = BASEDATETIME, tz = tz)
  res
}

interpolate_factor_element <- function(data, group, frame, ease) {
  all_levels <- levels(data)
  ord <- is.ordered(data)
  res <- interpolate_character_element(data, group, frame, ease)
  res[['data']] <- if (ord) ordered(res[['data']], all_levels) else factor(res[['data']], all_levels)
  res
}

interpolate_list_element <- function(data, group, frame, ease) {
  new_data <- list_element_interpolator(as.list(data), as.integer(group), as.integer(frame), as.character(ease))
  attributes(new_data$data) <- attributes(data)
  new_data
}

interpolate_numlist_element <- function(data, group, frame, ease) {
  new_data <- numlist_element_interpolator(lapply(data, as.numeric), as.integer(group), as.integer(frame), as.character(ease))
  attributes(new_data$data) <- attributes(data)
  new_data
}

get_phase_element <- function(data, group, frame, ease) {
  phase_element_interpolator(as.character(data), as.integer(group), as.integer(frame), as.character(ease))
}

#' Generator internals
#'
#' @param x A generator object
#'
#' @return Various data
#'
#' @name gen_internal
#' @rdname gen_internal
#'
#' @keywords internal
#'
NULL

#' @rdname gen_internal
#' @export
PHASE_LEVELS <- c("raw", "static", "transition", "enter", "exit")

#' @rdname gen_internal
#' @export
is_generator <- function(x) inherits(x, "frame_generator")

#' @rdname gen_internal
#' @export
generator_settings <- function(x) attr(x, "generator_settings")

#' @rdname gen_internal
#' @export
`generator_settings<-` <- function(x, value) {
  attr(x, "generator_settings") <- value
  x
}

#' @rdname gen_internal
#' @export
ease_type <- function(x) generator_settings(x)$ease_type

#' @rdname gen_internal
#' @export
`ease_type<-` <- function(x, value) {
  generator_settings(x)$ease_type <- value
  x
}

#' @rdname gen_internal
#' @export
col_types <- function(x) generator_settings(x)$col_types

#' @rdname gen_internal
#' @export
`col_types<-` <- function(x, value) {
  generator_settings(x)$col_types <- value
  x
}

#' @rdname gen_internal
#' @export
gen_data <- function(x) generator_settings(x)$data

#' @rdname gen_internal
#' @export
`gen_data<-` <- function(x, value) {
  generator_settings(x)$data <- value
  x
}

convert_generator <- function(x) {
  UseMethod('convert_generator')
}
#' @export
convert_generator.default <- function(x) {
  list(data = x, settings = list(converter = function(x, ...) x))
}

#' @rdname gen_internal
#' @export
gen_to_data_frame <- function(...) {
  data <- lapply(list(...), convert_generator)
  settings <- lapply(data, `[[`, 'settings')
  data <- lapply(data, `[[`, 'data')
  id <- rep(seq_along(data), vapply(data, nrow, integer(1)))
  data <- vec_rbind(data)
  attr(data, "generator_id") <- id
  attr(data, "generator_info") <- settings
  class(data) <- c("generator_df", class(data))
  data
}

#' @rdname gen_internal
#' @export
data_frame_to_gen <- function(x) {
  if (!inherits(x, "generator_df")) {
    stop('This does not appear to be a generator in data frame disguise')
  }
  info <- attr(x, "generator_info")
  data <- split(data.frame(x), attr(x, "generator_id"))
  Map(function(data, info) {
    info$converter(data, info)
  })
}

#' Generator for tweening components separately from each other
#'
#' This is a generator versions of [tween_components()]. It returns a generator
#' that can be used with [get_frame()] and [get_raw_frames()] to extract frames
#' for a specific time point scaled between 0 and 1.
#'
#' @inheritParams tween_components
#'
#' @return A `component_generator` object
#'
#' @family Other generators
#'
#' @export
#' @importFrom rlang eval_tidy enquo quo_is_null
#'
#' @examples
#' from_zero <- function(x) {x$x <- 0; x}
#'
#' data <- data.frame(
#'   x = c(1, 2, 2, 1, 2, 2),
#'   y = c(1, 2, 2, 2, 1, 1),
#'   time = c(1, 4, 8, 4, 8, 10),
#'   id = c(1, 1, 1, 2, 2, 2)
#' )
#'
#' gen <- gen_components(data, 'cubic-in-out', time = time, id = id,
#'                       enter = from_zero, enter_length = 4)
#'
#' get_frame(gen, 0.3)
gen_components <- function(.data, ease, nframes, time, id = NULL, range = NULL,
                           enter = NULL, exit = NULL, enter_length = 0,
                           exit_length = 0) {
  time <- enquo(time)
  time <- eval_tidy(time, .data)
  id <- enquo(id)
  id <- if (quo_is_null(id)) rep(1, nrow(.data)) else eval_tidy(id, .data)
  if (is.null(enter_length)) enter_length <- 0
  if (is.null(exit_length)) exit_length <- 0

  .data$.phase <- NULL
  if (length(ease) == 1) ease <- rep(ease, ncol(.data))
  if (length(ease) == ncol(.data)) {
    ease <- c(ease, 'linear') # To account for .phase column
  } else {
    stop('Ease must be either a single string or one for each column', call. = FALSE)
  }

  .data$.phase <- rep_len(factor("raw", levels = PHASE_LEVELS), nrow(.data))
  class(.data) <- c(c("component_generator", "frame_generator"), class(.data))

  gen_data <- .complete_components(.data, time, id, enter, exit, enter_length, exit_length)
  time <- gen_data$.time
  id <- gen_data$.id
  gen_data$.time <- NULL
  gen_data$.id <- NULL
  d_order <- order(id, time)
  if (is.null(range)) range <- range(time)
  if (diff(range) == 0) stop('range cannot be 0', call. = FALSE)

  generator_settings(.data) <- list(
    data = gen_data[d_order, ],
    id = id[d_order],
    time = time[d_order],
    range = range,
    ease_type = ease,
    col_types = col_classes(.data)
  )

  .data
}

#' @export
get_frame.component_generator <- function(generator, at, ...) {
  d <- generator_settings(generator)
  # clamp between 0 and 1
  at <- min(max(at, 0), 1)
  range <- d$range

  # normalise to range
  at <- (range[2] - range[1]) * at + range[1]

  data <- gen_data(generator)
  ease <- ease_type(generator)
  type <- col_types(generator)

  frame <- lapply(seq_along(data), function(i) {
    col <- data[[i]]
    e <- rep_len(ease[[i]], length(col))
    switch(
      type[i],
      numeric = interpolate_numeric_element_at(col, d$id, d$time, at, e),
      logical = interpolate_logical_element_at(col, d$id, d$time, at, e),
      factor = interpolate_factor_element_at(col, d$id, d$time, at, e),
      character = interpolate_character_element_at(col, d$id, d$time, at, e),
      colour = interpolate_colour_element_at(col, d$id, d$time, at, e),
      date = interpolate_date_element_at(col, d$id, d$time, at, e),
      datetime = interpolate_datetime_element_at(col, d$id, d$time, at, e),
      constant = interpolate_constant_element_at(col, d$id, d$time, at, e),
      numlist = interpolate_numlist_element_at(col, d$id, d$time, at, e),
      list = interpolate_list_element_at(col, d$id, d$time, at, e),
      phase = get_phase_element_at(col, d$id, d$time, at, e)
    )
  })
  structure(frame, names = names(data), row.names = .set_row_names(length(frame[[1]])), class = 'data.frame')
}
#' @export
get_raw_frames.component_generator <- function(generator, at, before = 0, after = 0, ...) {
  d <- generator_settings(generator)

  # clamp between 0 and 1
  before <- min(max(at - before, 0), 1)
  after <- min(max(at + after, 0), 1)
  at <- min(max(at, 0), 1)
  range <- d$range

  # normalise to generator time
  at <- (range[2] - range[1]) * at + range[1]
  before <- (range[2] - range[1]) * before + range[1]
  after <- (range[2] - range[1]) * after + range[1]

  # Find before and after
  before <- d$time >= before & d$time < at
  after <- d$time > at & d$time <= after
  raw <- d$data$.phase == 'raw'

  list(
    before = d$data[before & raw, , drop = FALSE],
    after = d$data[after & raw, , drop = FALSE]
  )
}


#' @export
convert_generator.component_generator <- convert_generator.along_generator

#' @rdname tween
#'
#' @export
tween_datetime <- function(data, n, ease = 'linear') {
  data <- as.list(data)
  prepData <- prepareTween(data, n, ease)
  if (!all(sapply(prepData$data, inherits, what = 'POSIXt'))) {
    stop('data must consist of POSIXt elements')
  }
  tweendata <- do.call(interpolate_datetime_state, prepData)
  unname(split(tweendata,
               rep(seq_along(data[[1]]), length.out = length(tweendata))))
}

#' @rdname tween
#'
#' @export
tween_datetime_t <- function(data, n, ease = 'linear') {
  if (!is.list(data)) {
    data <- list(data)
  }
  prepData <- prepareTweenTranspose(data, n, ease)
  if (!all(sapply(prepData$data, inherits, what = 'POSIXt'))) {
    stop('data must consist of POSIXt elements')
  }
  tweendata <- do.call(interpolate_datetime_state, prepData)
  unname(split(tweendata,
               rep(seq_along(data), rep(n, length.out = length(data)))))
}

interpolate_numeric_at <- function(from, to, at, ease) {
  numeric_at_interpolator(as.numeric(from), as.numeric(to), as.numeric(at), as.character(ease))
}

interpolate_logical_at <- function(from, to, at, ease) {
  as.logical(round(interpolate_numeric_at(from, to, at, ease)))
}

#' @importFrom farver decode_colour encode_colour
interpolate_colour_at <- function(from, to, at, ease) {
  from <- decode_colour(from, alpha = TRUE, to = 'lab')
  to <- decode_colour(to, alpha = TRUE, to = 'lab')
  data <- colour_at_interpolator(from, to, as.numeric(at), as.character(ease))
  encode_colour(data[, 1:3, drop = FALSE], alpha = data[,4], from = 'lab')
}

interpolate_constant_at <- function(from, to, at, ease) {
  constant_at_interpolator(as.character(from), as.character(to), as.numeric(at), as.character(ease))
}

interpolate_character_at <- interpolate_constant_at

interpolate_date_at <- function(from, to, at, ease) {
  data <- interpolate_numeric_at(from, to, at, ease)
  as.Date(data, origin = BASEDATE)
}

interpolate_datetime_at <- function(from, to, at, ease) {
  if (inherits(from, 'POSIXlt')) {
    warning("POSIXlt converted to POSIXct")
    from <- as.POSIXct(from)
  }
  tz <- attr(from, 'tzone')
  data <- interpolate_numeric_at(from, to, at, ease)
  as.POSIXct(data, origin = BASEDATETIME, tz = tz)
}

interpolate_factor_at <- function(from, to, at, ease) {
  all_levels <- unique(c(levels(from), levels(to)))
  data <- interpolate_constant_at(from, to, at, ease)
  if (is.ordered(from)) ordered(data, all_levels) else factor(data, all_levels)
}

interpolate_list_at <- function(from, to, at, ease) {
  data <- list_at_interpolator(as.list(from), as.list(to), as.numeric(at), as.character(ease))
  attributes(data) <- attributes(from)
  data
}

interpolate_numlist_at <- function(from, to, at, ease) {
  data <- numlist_at_interpolator(lapply(from, as.numeric), lapply(to, as.numeric), as.numeric(at), as.character(ease))
  attributes(data) <- attributes(from)
  data
}

interpolate_numeric_at_t <- function(from, to, at, ease) {
  numeric_at_t_interpolator(as.numeric(from), as.numeric(to), as.numeric(at), as.character(ease))
}

interpolate_logical_at_t <- function(from, to, at, ease) {
  as.logical(round(interpolate_numeric_at_t(from, to, at, ease)))
}

#' @importFrom farver decode_colour encode_colour
interpolate_colour_at_t <- function(from, to, at, ease) {
  from <- decode_colour(from, alpha = TRUE, to = 'lab')
  to <- decode_colour(to, alpha = TRUE, to = 'lab')
  data <- colour_at_t_interpolator(from, to, as.numeric(at), as.character(ease))
  encode_colour(data[, 1:3, drop = FALSE], alpha = data[,4], from = 'lab')
}

interpolate_constant_at_t <- function(from, to, at, ease) {
  constant_at_t_interpolator(as.character(from), as.character(to), as.numeric(at), as.character(ease))
}

interpolate_character_at_t <- interpolate_constant_at_t

interpolate_date_at_t <- function(from, to, at, ease) {
  data <- interpolate_numeric_at(from, to, at, ease)
  as.Date(data, origin = BASEDATE)
}

interpolate_datetime_at_t <- function(from, to, at, ease) {
  if (inherits(from, 'POSIXlt')) {
    warning("POSIXlt converted to POSIXct")
    from <- as.POSIXct(from)
  }
  tz <- attr(from, 'tzone')
  data <- interpolate_numeric_at(from, to, at, ease)
  as.POSIXct(data, origin = BASEDATETIME, tz = tz)
}

interpolate_factor_at_t <- function(from, to, at, ease) {
  all_levels <- unique(c(levels(from), levels(to)))
  data <- interpolate_constant_at(from, to, at, ease)
  if (is.ordered(from)) ordered(data, all_levels) else factor(data, all_levels)
}

interpolate_list_at_t <- function(from, to, at, ease) {
  data <- list_at_t_interpolator(as.list(from), as.list(to), as.numeric(at), as.character(ease))
  attributes(data) <- attributes(from)
  data
}

interpolate_numlist_at_t <- function(from, to, at, ease) {
  data <- numlist_at_t_interpolator(lapply(from, as.numeric), lapply(to, as.numeric), as.numeric(at), as.character(ease))
  attributes(data) <- attributes(from)
  data
}

interpolate_numeric_element_at <- function(data, group, time, at, ease) {
  numeric_element_at_interpolator(as.numeric(data), as.integer(group), as.numeric(time), as.numeric(at), as.character(ease))
}

interpolate_logical_element_at <- function(data, group, time, at, ease) {
  as.logical(interpolate_numeric_element_at(data, group, time, at, ease))
}

#' @importFrom farver decode_colour encode_colour
interpolate_colour_element_at <- function(data, group, time, at, ease) {
  data <- decode_colour(data, alpha = TRUE, to = 'lab')
  col <- colour_element_at_interpolator(data, as.integer(group), as.numeric(time), as.numeric(at), as.character(ease))
  encode_colour(col[, 1:3, drop = FALSE], alpha = col[,4], from = 'lab')
}

interpolate_constant_element_at <- function(data, group, time, at, ease) {
  constant_element_at_interpolator(as.character(data), as.integer(group), as.numeric(time), as.numeric(at), as.character(ease))
}

interpolate_character_element_at <- interpolate_constant_element_at

interpolate_date_element_at <- function(data, group, time, at, ease) {
  res <- interpolate_numeric_element_at(data, group, time, at, ease)
  as.Date(res, origin = BASEDATE)
}

interpolate_datetime_element_at <- function(data, group, time, at, ease) {
  if (inherits(data, 'POSIXlt')) {
    warning("POSIXlt converted to POSIXct")
    data <- as.POSIXct(data)
  }
  tz <- attr(data, 'tzone')
  res <- interpolate_numeric_element_at(data, group, time, at, ease)
  as.POSIXct(res, origin = BASEDATETIME, tz = tz)
}

interpolate_factor_element_at <- function(data, group, time, at, ease) {
  all_levels <- levels(data)
  ord <- is.ordered(data)
  res <- interpolate_character_element_at(data, group, time, at, ease)
  if (ord) ordered(res[['data']], all_levels) else factor(res[['data']], all_levels)
}

interpolate_list_element_at <- function(data, group, time, at, ease) {
  new_data <- list_element_at_interpolator(as.list(data), as.integer(group), as.numeric(time), as.numeric(at), as.character(ease))
  attributes(new_data) <- attributes(data)
  new_data
}

interpolate_numlist_element_at <- function(data, group, time, at, ease) {
  new_data <- numlist_element_at_interpolator(lapply(data, as.numeric), as.integer(group), as.numeric(time), as.numeric(at), as.character(ease))
  attributes(new_data) <- attributes(data)
  new_data
}

get_phase_element_at <- function(data, group, time, at, ease) {
  phase_element_at_interpolator(as.character(data), as.integer(group), as.numeric(time), as.numeric(at), as.character(ease))
}

#' @rdname tween
#'
#' @export
tween_colour <- function(data, n, ease = 'linear') {
  data <- as.list(data)
  prepData <- prepareTween(data, n, ease)
  if (anyNA(suppressWarnings(col2rgb(unlist(prepData$data))))) {
    stop('all elements in data must be convertible to colour')
  }
  tweendata <- do.call(interpolate_colour_state, prepData)
  unname(split(tweendata,
               rep(seq_along(data[[1]]), length.out = length(tweendata))))
}
#' @rdname tween
#'
#' @export
tween_color <- tween_colour

#' @rdname tween
#'
#' @export
tween_colour_t <- function(data, n, ease = 'linear') {
  if (!is.list(data)) {
    data <- list(data)
  }
  prepData <- prepareTweenTranspose(data, n, ease)
  if (anyNA(suppressWarnings(col2rgb(unlist(prepData$data))))) {
    stop('all elements in data must be convertible to colour')
  }
  tweendata <- do.call(interpolate_colour_state, prepData)
  unname(split(tweendata,
               rep(seq_along(data), rep(n, length.out = length(data)))))
}
#' @rdname tween
#'
#' @export
tween_color_t <- tween_colour_t

#' Interpolate data along a given dimension
#'
#' This tween takes groups of rows along with the time for each row and
#' calculates the exact value at each at each frame. Further it allows for
#' keeping the subsequent raw data from previous frame as well as letting the
#' final row linger beyond its time. It especially useful for data that should
#' be visualised as lines that are drawn along the x-axis, but can of course
#' also be used for other dimensions as well (even dimensions not corresponding
#' to any axis).
#'
#' @inheritParams tween_components
#' @param along The "time" point for each row
#' @param history Should earlier datapoints be kept in subsequent frames
#' @param keep_last Should the last point of each id be kept beyond its time
#'
#' @return A data.frame with the same columns as `.data` along with `.id` giving
#' the component id, `.phase` giving the state of each component in each frame,
#' and `.frame` giving the frame membership of each row.
#'
#' @family data.frame tween
#'
#' @importFrom rlang enquo quo_is_null eval_tidy is_integerish
#' @export
tween_along <- function(.data, ease, nframes, along, id = NULL, range = NULL, history = TRUE, keep_last = FALSE) {
  along <- enquo(along)
  along <- as.numeric(eval_tidy(along, .data))
  id <- enquo(id)
  id <- if (quo_is_null(id)) rep(1, nrow(.data)) else eval_tidy(id, .data)
  .data <- .complete_along(.data, along, id)

  if (length(ease) == 1) ease <- rep(ease, ncol(.data) - 3)
  if (length(ease) == ncol(.data) - 3) {
    ease <- c(ease, 'linear', 'linear', 'linear') # To account for .phase and .id columns
  } else {
    stop('Ease must be either a single string or one for each column', call. = FALSE)
  }
  if (!is_integerish(nframes, 1L)) {
    stop("`nframes` must be a single count", call. = FALSE)
  }

  timerange <- if (is.null(range)) range(.data$.time) else range
  timerange <- as.numeric(timerange)
  if (diff(timerange) == 0) stop('range must have a length', call. = FALSE)
  framelength <- diff(timerange) / (nframes - 1)
  frame <- 1 + (nframes - 1) * (.data$.time - timerange[1]) / diff(timerange)
  frames <- seq_len(nframes)
  groups <- unique(.data$.id)
  group <- match(.data$.id, groups)
  colClasses <- col_classes(.data)
  tweendata <- lapply(seq_along(.data),  function(i) {
    d <- .data[[i]]
    e <- ease[i]
    switch(
      colClasses[i],
      numeric = interpolate_numeric_along(d, group, frame, frames, e, history, keep_last),
      logical = interpolate_logical_along(d, group, frame, frames, e, history, keep_last),
      factor = interpolate_factor_along(d, group, frame, frames, e, history, keep_last),
      character = interpolate_character_along(d, group, frame, frames, e, history, keep_last),
      colour = interpolate_colour_along(d, group, frame, frames, e, history, keep_last),
      date = interpolate_date_along(d, group, frame, frames, e, history, keep_last),
      datetime = interpolate_datetime_along(d, group, frame, frames, e, history, keep_last),
      constant = interpolate_constant_along(d, group, frame, frames, e, history, keep_last),
      numlist = interpolate_numlist_along(d, group, frame, frames, e, history, keep_last),
      list = interpolate_list_along(d, group, frame, frames, e, history, keep_last),
      phase = get_phase_along(group, frame, frames, history, keep_last)
    )
  })
  tweenInfo <- tweendata[[1]][, c('group', 'frame')]
  tweendata <- lapply(tweendata, `[[`, i = 'data')
  tweendata <- structure(tweendata, names = names(.data), row.names = .set_row_names(length(tweendata[[1]])), class = 'data.frame')
  tweendata$.frame <- tweenInfo$frame
  tweendata$.id <- tweenInfo$group
  tweendata$.time <- unsplit(
    lapply(split(tweendata$.time, tweendata$.frame), function(x) rlang::rep_along(x, max(x))),
    tweendata$.frame
  )
  attr(tweendata, 'framelength') <- framelength
  tweendata[order(tweendata$.frame, tweendata$.id), , drop = FALSE]
}

.complete_along <- function(data, along, id) {
  if (length(along) != nrow(data) || length(id) != nrow(data)) {
    stop('along and id must be the same length as the number of rows in data', call. = FALSE)
  }
  data <- data[order(id), , drop = FALSE]
  along <- along[order(id)]
  id <- sort(id)
  data$.id <- id
  data$.phase <- 'raw'
  data$.time <- along
  data
}

context("along")

df <- data.frame(
  x = c(1, 5, 7, 10),
  y = c(4, 3, 7, -1),
  col = c('black', 'red', 'green', 'blue'),
  type = letters[1:4],
  stringsAsFactors = FALSE
)

test_that("tween_along works", {
  tween <- tween_along(df, ease = 'linear', nframes = 10, along = x)
  expect_equal(nrow(tween), 30)
  expect_equal(tween$col[22], '#78B785')
  expect_equal(tween$y[8], 3.25)

  tween <- tween_along(df, ease = 'linear', nframes = 10, along = x, history = FALSE)
  expect_equal(nrow(tween), 9)
  expect_equal(tween$col[8], '#78B785')
  expect_equal(tween$y[2], 3.75)
})

test_that("tween_along throws errors", {
  expect_error(tween_along(df, ease = 'linear', nframes = 10, along = 1))
  expect_error(tween_along(df, ease = 'linear', nframes = 10, along = x, id = 1))
  expect_error(tween_along(df, ease = 'linear', nframes = 10, along = x, range = c(0, 0)))
  expect_error(tween_along(df[1,], ease = 'linear', nframes = 10, along = x))
})

#' Get a specific position between two states
#'
#' This tween allows you to query a specific postion between two states rather
#' than generate evenly spaced states. It can work with either data.frames or
#' single vectors and each row/element can have its own position and easing.
#'
#' @param from,to A data.frame or vector of the same type. If either is of
#' length/nrow 1 it will get repeated to match the length of the other
#' @param at A numeric between 0 and 1 recycled to match the nrow/length of
#' `from`
#' @param ease A character vector giving valid easing functions. Recycled to
#' match the ncol of `from`
#'
#' @return If `from`/`to` is a data.frame then a data.frame with the same
#' columns. If `from`/`to` is a vector then a vector.
#'
#' @export
#'
#' @examples
#' tween_at(mtcars[1:6, ], mtcars[6:1, ], runif(6), 'cubic-in-out')
#'
tween_at <- function(from, to, at, ease) {
  single_vec <- !is.data.frame(from)
  if (single_vec) {
    if (length(from) == 0 || length(to) == 0) return(to[integer()])
    from_df <- data.frame(data = rep(NA, length(from)))
    to_df <- data.frame(data = rep(NA, length(to)))
    from_df$data <- from
    to_df$data <- to
    from <- from_df
    to <- to_df
  } else {
    if (nrow(from) == 0 || nrow(to) == 0) return(to[integer(), ])
  }
  if (length(at) == 0) stop('at must have length > 0', call. = FALSE)
  if (nrow(from) == 1) from <- from[rep(1, nrow(to)), , drop = FALSE]
  if (nrow(to) == 1) to <- to[rep(1, nrow(from)), , drop = FALSE]
  if (nrow(from) != nrow(to)) {
    stop('from and to must be same length', call. = FALSE)
  }
  if (any(names(from) != names(to))) {
    stop('`from` and `to` must have the same columns', call. = FALSE)
  }
  at <- rep(at, length.out = nrow(from))
  ease <- rep(ease, length.out = ncol(from))
  classes <- col_classes(from)
  to_classes <- col_classes(to)
  mismatch <- to_classes != classes
  for (i in which(mismatch)) {
    all_na_to <- all(is.na(to[[i]]))
    all_na_from <- all(is.na(from[[i]]))
    if (all_na_from) {
      storage.mode(from[[i]]) <- storage.mode(to[[i]])
    } else if (all_na_to) {
      storage.mode(to[[i]]) <- storage.mode(from[[i]])
    } else {
      stop('The ', names(to)[i], 'column differs in type between the two inputs', call. = FALSE)
    }
  }
  tweendata <- lapply(seq_along(classes), function(i) {
    switch(
      classes[i],
      numeric = interpolate_numeric_at(from[[i]], to[[i]], at, ease[i]),
      logical = interpolate_logical_at(from[[i]], to[[i]], at, ease[i]),
      factor = interpolate_factor_at(from[[i]], to[[i]], at, ease[i]),
      character = interpolate_character_at(from[[i]], to[[i]], at, ease[i]),
      colour = interpolate_colour_at(from[[i]], to[[i]], at, ease[i]),
      date = interpolate_date_at(from[[i]], to[[i]], at, ease[i]),
      datetime = interpolate_datetime_at(from[[i]], to[[i]], at, ease[i]),
      constant = interpolate_constant_at(from[[i]], to[[i]], at, ease[i]),
      numlist = interpolate_numlist_at(from[[i]], to[[i]], at, ease[i]),
      list = interpolate_list_at(from[[i]], to[[i]], at, ease[i]),
      phase = ifelse(from[[i]] == "enter", "enter", ifelse(to[[i]] == "exit", "exit", "transition"))
    )
  })
  if (single_vec) return(tweendata[[1]])

  structure(tweendata, names = names(from), row.names = .set_row_names(length(tweendata[[1]])), class = 'data.frame')
}
#' Get several specific position between two states
#'
#' This tween is a variation of [tween_at()]. Instead of having `at` refer to
#' the tweening position of each row, each `at` will interpolate the full data
#' at that position.
#'
#' @param from,to A data.frame or vector of the same type. If either is of
#' length/nrow 1 it will get repeated to match the length of the other
#' @param at A numeric vector with values between 0 and 1.
#' @param ease A character vector giving valid easing functions. Recycled to
#' match the ncol of `from`
#'
#' @return If `from`/`to` is a data.frame then a data.frame with the same
#' columns. If `from`/`to` is a vector then a vector.
#'
#' @export
#'
#' @examples
#' tween_at_t(mtcars[1:6, ], mtcars[6:1, ], runif(3), 'cubic-in-out')
#'
tween_at_t <- function(from, to, at, ease) {
  single_vec <- !is.data.frame(from)
  if (single_vec) {
    if (length(from) == 0 || length(to) == 0) return(to[integer()])
    from_df <- data.frame(data = rep(NA, length(from)))
    to_df <- data.frame(data = rep(NA, length(to)))
    from_df$data <- from
    to_df$data <- to
    from <- from_df
    to <- to_df
  } else {
    if (nrow(from) == 0 || nrow(to) == 0) return(to[integer(), ])
  }
  if (length(at) == 0) stop('at must have length > 0', call. = FALSE)
  if (nrow(from) == 1) from <- from[rep(1, nrow(to)), , drop = FALSE]
  if (nrow(to) == 1) to <- to[rep(1, nrow(from)), , drop = FALSE]
  if (nrow(from) != nrow(to)) {
    stop('from and to must be same length', call. = FALSE)
  }
  if (any(names(from) != names(to))) {
    stop('`from` and `to` must have the same columns', call. = FALSE)
  }
  ease <- rep(ease, length.out = ncol(from))
  classes <- col_classes(from)
  to_classes <- col_classes(to)
  mismatch <- to_classes != classes
  for (i in which(mismatch)) {
    all_na_to <- all(is.na(to[[i]]))
    all_na_from <- all(is.na(from[[i]]))
    if (all_na_from) {
      storage.mode(from[[i]]) <- storage.mode(to[[i]])
    } else if (all_na_to) {
      storage.mode(to[[i]]) <- storage.mode(from[[i]])
    } else {
      stop('The ', names(to)[i], 'column differs in type between the two inputs', call. = FALSE)
    }
  }
  tweendata <- lapply(seq_along(classes), function(i) {
    switch(
      classes[i],
      numeric = interpolate_numeric_at_t(from[[i]], to[[i]], at, ease[i]),
      logical = interpolate_logical_at_t(from[[i]], to[[i]], at, ease[i]),
      factor = interpolate_factor_at_t(from[[i]], to[[i]], at, ease[i]),
      character = interpolate_character_at_t(from[[i]], to[[i]], at, ease[i]),
      colour = interpolate_colour_at_t(from[[i]], to[[i]], at, ease[i]),
      date = interpolate_date_at_t(from[[i]], to[[i]], at, ease[i]),
      datetime = interpolate_datetime_at_t(from[[i]], to[[i]], at, ease[i]),
      constant = interpolate_constant_at_t(from[[i]], to[[i]], at, ease[i]),
      numlist = interpolate_numlist_at_t(from[[i]], to[[i]], at, ease[i]),
      list = interpolate_list_at_t(from[[i]], to[[i]], at, ease[i]),
      phase = interpolate_phase_at_t(from[[i]], to[[i]], at)
    )
  })
  if (single_vec) return(tweendata[[1]])

  tweendata$.frame <- rep(seq_along(at), each = length(tweendata[[1]]))

  structure(tweendata, names = names(from), row.names = .set_row_names(length(tweendata[[1]])), class = 'data.frame')
}

interpolate_phase_at_t <- function(from, to, at) {
  phase <- rep(ifelse(from == "enter", "enter", ifelse(to == "exit", "exit", "transition")), times = length(at))
  start_or_end <- at %in% c(0, 1)
  if (any(start_or_end)) {
    start_or_end <- rep(start_or_end, each = length(from))
    phase[phase == "transition" & start_or_end] <- "raw"
  }
  phase
}

context("at")

df1 <- data.frame(x = 1:2, y = 4:5, col = 'black', type = letters[1:2], stringsAsFactors = FALSE)
df2 <- data.frame(x = 11:12, y = 14:15, col = 'white', type = letters[1], stringsAsFactors = FALSE)

test_that("tween_at works", {
  tween <- tween_at(df1, df2, 0.5, 'linear')
  expect_equal(nrow(tween), 2)
  expect_named(tween, names(df1))
  expect_equal(tween$x, c(6, 7))
  expect_equal(tween$col[1], '#777777')
})

test_that("tween_at handles weird input", {
  tween <- tween_at(df1, df2[1,], 0.5, 'linear')
  expect_equal(nrow(tween), 2)
  tween <- tween_at(df1[1,], df2, 0.5, 'linear')
  expect_equal(nrow(tween), 2)
  tween <- tween_at(df1, df2[integer(),], 0.5, 'linear')
  expect_equal(nrow(tween), 0)
  tween <- tween_at(df1[integer(),], df2, 0.5, 'linear')
  expect_equal(nrow(tween), 0)
  expect_error(tween_at(df1[c(1,2,1), ], df2, 0.5, 'linear'))
  expect_error(tween_at(df1, df2, numeric(), 'linear'))
  expect_error(tween_at(df1, df2, 0.5, character()))
})

test_that('tween_at works with vectors', {
  tween <- tween_at(df1$x, df2$x, 0.5, 'linear')
  expect_is(tween, 'numeric')
  expect_equal(tween, c(6,7))
  expect_error(tween_at(df1$x, df2$col))
})

#' Interpolate individual component
#'
#' This function is much like [tween_elements()] but with a slightly different
#' syntax and support for many of the newer features such as enter/exits and
#' tween phase identification. Furthermore it uses tidy evaluation for time and
#' id, making it easier to change these on the fly. The biggest change in terms
#' of functionality compared to `tween_elements()` is that the easing function
#' is now given per column and not per row. If different easing functions are
#' needed for each transition then `tween_elements()` is needed.
#'
#' @inheritParams tween_state
#'
#' @param .data A data.frame with components at different stages
#'
#' @param time An unquoted expression giving the timepoint for the different
#' stages of the components. Will be evaluated in the context of `.data` so can
#' refer to a column from that
#'
#' @param id An unquoted expression giving the component id for each row. Will
#' be evaluated in the context of `.data` so can refer to a column from that
#'
#' @param range The range of time points to include in the tween. If `NULL` it
#' will use the range of `time`
#'
#' @param enter_length,exit_length The lenght of the opening and closing
#' transitions if `enter` and/or `exit` is given. Measured in the same units as
#' `time`
#'
#' @return A data.frame with the same columns as `.data` along with `.id` giving
#' the component id, `.phase` giving the state of each component in each frame,
#' and `.frame` giving the frame membership of each row.
#'
#' @family data.frame tween
#'
#' @examples
#'
#' from_zero <- function(x) {x$x <- 0; x}
#'
#' data <- data.frame(
#'   x = c(1, 2, 2, 1, 2, 2),
#'   y = c(1, 2, 2, 2, 1, 1),
#'   time = c(1, 4, 10, 4, 8, 10),
#'   id = c(1, 1, 1, 2, 2, 2)
#' )
#'
#' data <- tween_components(data, 'cubic-in-out', nframes = 100, time = time,
#'                          id = id, enter = from_zero, enter_length = 4)
#'
#' @export
#' @importFrom rlang enquo eval_tidy
#'
tween_components <- function(.data, ease, nframes, time, id = NULL, range = NULL, enter = NULL, exit = NULL, enter_length = 0, exit_length = 0) {
  time <- enquo(time)
  time <- eval_tidy(time, .data)
  id <- enquo(id)
  id <- if (quo_is_null(id)) rep(1, nrow(.data)) else eval_tidy(id, .data)
  if (is.null(enter_length)) enter_length <- 0
  if (is.null(exit_length)) exit_length <- 0
  .data <- .complete_components(.data, time, id, enter, exit, enter_length, exit_length)

  .tween_individuals(.data, ease, nframes, range)
}

.tween_individuals <- function(.data, ease, nframes, range) {
  if (nframes == 0) return(.data[integer(), , drop = FALSE])
  if (nrow(.data) == 0) return(.data)
  if (length(ease) == 1) ease <- rep(ease, ncol(.data) - 3)
  if (length(ease) == ncol(.data) - 3) {
    ease <- c(ease, 'linear', 'linear', 'linear') # To account for .phase and .id columns
  } else {
    stop('Ease must be either a single string or one for each column', call. = FALSE)
  }
  if (!is_integerish(nframes, 1L)) {
    stop("`nframes` must be a single count", call. = FALSE)
  }

  timerange <- if (is.null(range)) range(.data$.time) else range
  if (diff(timerange) == 0) stop('range must have a length', call. = FALSE)
  framelength <- diff(timerange) / (nframes - 1)
  .data <- .data[order(.data$.id, .data$.time), , drop = FALSE]
  frame <- round((.data$.time - min(timerange[1])) / framelength) + 1
  .data$.time <- NULL
  colClasses <- col_classes(.data)
  tweendata <- lapply(seq_along(.data),  function(i) {
    d <- .data[[i]]
    e <- rep(ease[i], length(d))
    switch(
      colClasses[i],
      numeric = interpolate_numeric_element(d, .data$.id, frame, e),
      logical = interpolate_logical_element(d, .data$.id, frame, e),
      factor = interpolate_factor_element(d, .data$.id, frame, e),
      character = interpolate_character_element(d, .data$.id, frame, e),
      colour = interpolate_colour_element(d, .data$.id, frame, e),
      date = interpolate_date_element(d, .data$.id, frame, e),
      datetime = interpolate_datetime_element(d, .data$.id, frame, e),
      constant = interpolate_constant_element(d, .data$.id, frame, e),
      numlist = interpolate_numlist_element(d, .data$.id, frame, e),
      list = interpolate_list_element(d, .data$.id, frame, e),
      phase = get_phase_element(d, .data$.id, frame, e)
    )
  })
  tweenInfo <- tweendata[[1]][, c('group', 'frame')]
  tweendata <- lapply(tweendata, `[[`, i = 'data')
  tweendata <- structure(tweendata, names = names(.data), row.names = .set_row_names(length(tweendata[[1]])), class = 'data.frame')
  tweendata$.frame <- tweenInfo$frame
  tweendata$.id <- tweenInfo$group
  tweendata <- tweendata[tweendata$.frame >= 1 & tweendata$.frame <= nframes, , drop = FALSE]
  attr(tweendata, 'framelength') <- framelength
  tweendata[order(tweendata$.frame, tweendata$.id), , drop = FALSE]
}

#' @importFrom vctrs vec_rbind
#' @importFrom rlang as_function
.complete_components <- function(data, time, id, enter, exit, enter_length, exit_length) {
  if (length(id) != nrow(data) || length(time) != nrow(data)) {
    stop('id and time must have the same length as the number of rows in data', call. = FALSE)
  }
  data$.id <- id
  data$.phase <- rep('raw', nrow(data))
  data$.time <- time
  if (any(!is.null(enter), !is.null(exit))) {
    time_ord <- order(time)
    if (!is.null(enter)) {
      enter_data <- as_function(enter)(data[time_ord[!duplicated(id[time_ord])], , drop = FALSE])
      enter_data$.phase <- 'enter'
      enter_data$.time <- enter_data$.time - enter_length
    } else {
      enter_data <- data[0, , drop = FALSE]
    }
    if (!is.null(exit)) {
      exit_data <- as_function(exit)(data[time_ord[!duplicated(id[time_ord], fromLast = TRUE)], , drop = FALSE])
      exit_data$.phase <- 'exit'
      exit_data$.time <- exit_data$.time + exit_length
    } else {
      exit_data <- data[0, , drop = FALSE]
    }
    data <- vec_rbind(enter_data, data, exit_data)
  }
  data
}

context("components")

df <- data.frame(
  x = c(1, 5, 7, 10),
  y = c(4, 3, 7, -1),
  col = c('black', 'red', 'green', 'blue'),
  type = letters[1:4],
  stringsAsFactors = FALSE
)

test_that("tween_components works", {
  tween <- tween_components(df, 'linear', nframes = 10, time = c(1, 7, 13, 20))
  expect_equal(nrow(tween), 10)
  expect_equal(tween$x[6], 19/3)
  expect_equal(tween$col[2], '#52170B')
  expect_equal(max(tween$.frame), 10)
  expect_true(all(tween$.phase[c(1,4,7,10)] == 'raw'))
  expect_true(all(tween$.phase[-c(1,4,7,10)] == 'transition'))

  tween <- tween_components(df, 'linear', nframes = 10, time = c(1, 7, 13, 20), rep(c(1,2), 2))
  expect_equal(nrow(tween), 14)
  expect_equal(tween$x[12], 25/3)
  expect_equal(tween$col[2], '#162A10')
  expect_equal(max(tween$.frame), 10)
  expect_true(all(tween$.phase[c(1,5,10,14)] == 'raw'))
  expect_true(all(tween$.phase[-c(1,5,10,14)] == 'transition'))
})

test_that("enter/exit works", {
  tween <- tween_components(df, 'linear', nframes = 20, time = c(1, 7, 13, 20), enter = function(df) {
    df$x <- 0
    df$col <- 'red'
    df
  }, enter_length = 3)
  expect_equal(nrow(tween), 20)
  expect_equal(tween$x[3], 2/3, tolerance = 1e-7)
  expect_equal(tween$col[2], '#A51B0B')
  expect_equal(max(tween$.frame), 20)
  expect_true(all(tween$.phase[1:3] == 'enter'))
})

test_that("weird input gets caught", {
  tween <- tween_components(df, 'linear', nframes = 0, time = c(1, 7, 13, 20))
  expect_equal(nrow(tween), 0)
  tween <- tween_components(df[integer(), ], 'linear', nframes = 10, time = numeric())
  expect_equal(nrow(tween), 0)
  expect_error(tween_components(df, 'linear', nframes = 10, time = 1))
  expect_error(tween_components(df, 'linear', nframes = 0, time = c(1, 7, 13, 20), id = 1))
})

#' Tween a data.frame of appearances
#'
#' This function is intended for use when you have a data.frame of events at
#' different time points. This could be the appearance of an observation for
#' example. This function replicates your data `nframes` times and
#' calculates the duration of each frame. At each frame each row is
#' assigned an age based on the progression of frames and the entry point of in
#' time for that row. A negative age means that the row has not appeared yet.
#'
#' @param data A data.frame to tween
#'
#' @param time The name of the column that holds the time dimension. This does
#' not need to hold time data in the strictest sence - any numerical type will
#' do
#'
#' @param timerange The range of time to create the tween for. If missing it
#' will defaults to the range of the time column
#'
#' @param nframes The number of frames to create for the tween. If missing it
#' will create a frame for each full unit in `timerange` (e.g.
#' `timerange = c(1, 10)` will give `nframes = 10`)
#'
#' @return A data.frame as `data` but repeated `nframes` times and
#' with the additional columns `.age` and `.frame`
#'
#' @family data.frame tween
#'
#' @importFrom vctrs vec_rbind
#'
#' @examples
#' data <- data.frame(
#'   x = rnorm(100),
#'   y = rnorm(100),
#'   time = sample(50, 100, replace = TRUE)
#' )
#'
#' data <- tween_appear(data, 'time', nframes = 200)
#'
#' @export
#'
tween_appear <- function(data, time, timerange, nframes) {
  if (missing(timerange) || is.null(timerange)) {
    timerange <- range(data[[time]])
  }
  if (missing(nframes) || is.null(nframes)) {
    nframes <- ceiling(diff(timerange) + 1)
  }
  framelength <- diff(timerange) / nframes
  frametimes <- seq(timerange[1], timerange[2], length.out = nframes)

  tweendata <- lapply(seq_along(frametimes), function(i) {
    data$.age <- frametimes[i] - data[[time]]
    data$.frame <- i
    data
  })
  tweendata <- do.call(vec_rbind, tweendata)
  attr(tweendata, 'framelength') <- framelength
  tweendata
}

#' Create simple tweens
#'
#' This set of functions can be used to interpolate between single data types,
#' i.e. data not part of data.frames but stored in vectors. All functions come
#' in two flavours: the standard and a *_t version. The standard reads the data
#' as a list of states, each tween matched element-wise from state to state. The
#' *_t version uses the transposed representation where each element is a vector
#' of states. The standard approach can be used when each tween has the same
#' number of states and you want to control the number of point in each state
#' transition. The latter is useful when each tween consists of different
#' numbers of states and/or you want to specify the total number of points for
#' each tween.
#'
#' @section Difference Between `tween_numeric` and `approx()`:
#' `tween_numeric` (and `tween_numeric_t`) is superficially equivalent to
#' [stats::approx()], but there are differences.
#' [stats::approx()] will create evenly spaced points, at the expense
#' of not including the actual points in the input, while the reverse is true
#' for `tween_numeric`. Apart from that `tween_numeric` of course supports easing
#' functions and is vectorized.
#'
#' @details
#' `tween` and `tween_t` are wrappers around the other functions that tries to guess
#' the type of input data and choose the appropriate tween function. Unless you
#' have data that could be understood as a colour but is in fact a character
#' vector it should be safe to use these wrappers. It is probably safer and more
#' verbose to use the explicit functions within package code as they circumvent
#' the type inference and checks whether the input data matches the tween
#' function.
#'
#' `tween_numeric` will provide a linear interpolation between the points based on
#' the sequence returned by the easing function. `tween_date` and `tween_datetime`
#' converts to numeric, produces the tweening, and converts back again.
#' `tween_colour` converts colours into Lab and does the interpolation there,
#' converting back to sRGB after the tweening is done. `tween_constant` is a
#' catchall that converts the input into character and interpolates by switching
#' between states halfway through the transition.
#'
#' The meaning of the `n` and `ease` arguments differs somewhat
#' between the standard and *_t versions of the functions. In the standard
#' function `n` and `ease` refers to the length and easing function of
#' each transition, being recycled if necessary to `length(data) - 1`. In
#' the *_t functions `n` and `ease` refers to the total length of each
#' tween and the easing function to be applied to all transition for each tween.
#' The will both be recycled to `length(data)`.
#'
#' @param data A list of vectors or a single vector. In the standard functions
#' each element in the list must be of equal length; for the *_t functions
#' lengths can differ. If a single vector is used it will be eqivalent to using
#' `as.list(data)` for the standard functions and `list(data)` for the
#' *_t functions.
#'
#' @param n The number of elements per transition or tween. See details
#'
#' @param ease The easing function to use for each transition or tween. See
#' details. Defaults to `'linear'`
#'
#' @return A list with an element for each tween. That means that the length of
#' the return is equal to the length of the elements in `data` for the
#' standard functions and equal to the length of `data` for the *_t
#' functions.
#'
#' @examples
#' tween_numeric(list(1:3, 10:8, c(20, 60, 30)), 10)
#'
#' tween_colour_t(list(colours()[1:4], colours()[1:2], colours()[25:100]), 100)
#'
#' @export
#'
tween <- function(data, n, ease = 'linear') {
  type <- guessType(data)
  switch(
    type,
    numeric = tween_numeric(data, n, ease),
    date = tween_date(data, n, ease),
    datetime = tween_datetime(data, n, ease),
    colour = tween_colour(data, n, ease),
    tween_constant_t(data, n, ease)
  )
}
#' @rdname tween
#' @export
tween_t <- function(data, n, ease = 'linear') {
  type <- guessType(data)
  switch(
    type,
    numeric = tween_numeric_t(data, n, ease),
    date = tween_date_t(data, n, ease),
    datetime = tween_datetime_t(data, n, ease),
    colour = tween_colour_t(data, n, ease),
    tween_constant_t(data, n, ease)
  )
}
guessType <- function(data) {
  data <- unlist(data)
  col_classes(list(data))
}

#' Generator for tweening the appearance of elements
#'
#' This is a generator version of [tween_events()]. It returns a generator
#' that can be used with [get_frame()] and [get_raw_frames()] to extract frames
#' for a specific time point scaled between 0 and 1.
#'
#' @inheritParams tween_events
#'
#' @return A `component_generator` object
#'
#' @family Other generators
#'
#' @export
#' @importFrom rlang enquo eval_tidy quo_is_missing
#'
#' @examples
#' d <- data.frame(
#'   x = runif(20),
#'   y = runif(20),
#'   time = runif(20),
#'   duration = runif(20, max = 0.1)
#' )
#' from_left <- function(x) {
#'   x$x <- -0.5
#'   x
#' }
#' to_right <- function(x) {
#'   x$x <- 1.5
#'   x
#' }
#'
#' gen <- gen_events(d, 'cubic-in-out', start = time, end = time + duration,
#'                   enter = from_left, exit = to_right, enter_length = 0.1,
#'                   exit_length = 0.05)
#'
#' get_frame(gen, 0.65)
#'
gen_events <- function(.data, ease, start, end = NULL, range = NULL, enter = NULL, exit = NULL, enter_length = 0, exit_length = 0) {
  start <- enquo(start)
  if (quo_is_missing(start)) stop('start must be provided', call. = FALSE)
  start <- eval_tidy(start, .data)
  end <- enquo(end)
  end <- eval_tidy(end, .data)
  enter_length <- enquo(enter_length)
  enter_length <- eval_tidy(enter_length, .data)
  exit_length <- enquo(exit_length)
  exit_length <- eval_tidy(exit_length, .data)

  if (is.null(enter_length)) enter_length <- 0
  if (is.null(exit_length)) exit_length <- 0

  .data$.phase <- NULL
  if (length(ease) == 1) ease <- rep(ease, ncol(.data))
  if (length(ease) == ncol(.data)) {
    ease <- c(ease, 'linear') # To account for .phase column
  } else {
    stop('Ease must be either a single string or one for each column', call. = FALSE)
  }

  .data$.phase <- rep_len(factor("raw", levels = PHASE_LEVELS), nrow(.data))
  class(.data) <- c(c("component_generator", "frame_generator"), class(.data))

  gen_data <- .complete_events(.data, start, end, enter, exit, enter_length, exit_length)

  time <- gen_data$.time
  id <- gen_data$.id
  gen_data$.time <- NULL
  gen_data$.id <- NULL
  d_order <- order(id, time)
  if (is.null(range)) range <- range(time)
  if (diff(range) == 0) stop('range cannot be 0', call. = FALSE)

  generator_settings(.data) <- list(
    data = gen_data[d_order, ],
    id = id[d_order],
    time = time[d_order],
    range = range,
    ease_type = ease,
    col_types = col_classes(.data)
  )

  .data
}

#' Fill out missing values by interpolation
#'
#' This tween fills out `NA` elements (or `NULL` elements if `data` is a list)
#' by interpolating between the prior and next non-missing values.
#'
#' @param data A data.frame or vector.
#' @param ease A character vector giving valid easing functions. Recycled to
#' match the ncol of `data`
#'
#' @return If `data` is a data.frame then a data.frame with the same
#' columns. If `data` is a vector then a vector.
#'
#' @export
#'
#' @examples
#' # Single vector
#' tween_fill(c(1, NA, NA, NA, NA, NA, 2, 6, NA, NA, NA, -2), 'cubic-in-out')
#'
#' # Data frame
#' tween_fill(mtcars[c(1, NA, NA, NA, NA, 4, NA, NA, NA, 10), ], 'cubic-in')
#'
tween_fill <- function(data, ease) {
  single_vec <- !is.data.frame(data)
  if (single_vec) {
    if (length(data) == 0) return(data[integer()])
    data_df <- data.frame(data = rep(NA, length(data)))
    data_df$data <- data
    data <- data_df
  } else {
    if (nrow(data) == 0) return(data[integer(), ])
  }
  ease <- rep(ease, length.out = ncol(data))
  classes <- col_classes(data)
  tweendata <- lapply(seq_along(classes), function(i) {
    switch(
      classes[i],
      numeric = interpolate_numeric_fill(data[[i]], ease[i]),
      logical = interpolate_logical_fill(data[[i]], ease[i]),
      factor = interpolate_factor_fill(data[[i]], ease[i]),
      character = interpolate_character_fill(data[[i]], ease[i]),
      colour = interpolate_colour_fill(data[[i]], ease[i]),
      date = interpolate_date_fill(data[[i]], ease[i]),
      datetime = interpolate_datetime_fill(data[[i]], ease[i]),
      constant = interpolate_constant_fill(data[[i]], ease[i]),
      numlist = interpolate_numlist_fill(data[[i]], ease[i]),
      list = interpolate_list_fill(data[[i]], ease[i])
    )
  })
  if (single_vec) return(tweendata[[1]])

  structure(tweendata, names = names(data), row.names = .set_row_names(length(tweendata[[1]])), class = 'data.frame')
}

context("fill")

df <- data.frame(
  x = c(1, NA, NA, NA, 6, 4, NA, NA, NA, NA, 20),
  type = c('a', NA, NA, NA, 'b', 'c', NA, NA, NA, NA, 'd'),
  col = c('red', NA, NA, NA, 'blue', 'green', NA, NA, NA, NA, 'black'),
  stringsAsFactors = FALSE
)
test_that("tween_fill works", {
  tween <- tween_fill(df, 'linear')
  expect_equal(dim(df), dim(tween))
  expect_equal(tween$x[3], 3.5)
  expect_equal(tween$col[10], '#183112')

  expect_equal(tween_fill(df$col, 'linear'), tween$col)

  tween <- tween_fill(df[-c(1, 11), 1], 'linear')
  expect_equal(tween, df$x[-c(1, 11)])
})

BASEDATE <- Sys.Date() - as.numeric(Sys.Date())
BASEDATETIME <- Sys.time() - as.numeric(Sys.time())

validEase <- c(
  "linear",
  "quadratic-in",
  "quadratic-out",
  "quadratic-in-out",
  "cubic-in",
  "cubic-out",
  "cubic-in-out",
  "quartic-in",
  "quartic-out",
  "quartic-in-out",
  "quintic-in",
  "quintic-out",
  "quintic-in-out",
  "sine-in",
  "sine-out",
  "sine-in-out",
  "circular-in",
  "circular-out",
  "circular-in-out",
  "exponential-in",
  "exponential-out",
  "exponential-in-out",
  "elastic-in",
  "elastic-out",
  "elastic-in-out",
  "back-in",
  "back-out",
  "back-in-out",
  "bounce-in",
  "bounce-out",
  "bounce-in-out"
)

#' @importFrom magrittr %>%
#' @export
magrittr::`%>%`

#' @rdname gen_internal
#' @export
#' @importFrom farver decode_colour
col_classes <- function(data) {
  classes <- vapply(data, vec_tween_class, character(1))
  names(classes) <- names(data)
  classes[names(classes) == '.phase'] <- 'phase'
  classes
}

#' Get the nominal class of a vector
#'
#' @param x a vector
#'
#' @export
#' @keywords internal
vec_tween_class <- function(x) {
  UseMethod('vec_tween_class')
}
#' @export
vec_tween_class.default <- function(x) 'constant'
#' @export
vec_tween_class.numeric <- function(x) 'numeric'
#' @export
vec_tween_class.logical <- function(x) 'logical'
#' @export
vec_tween_class.factor <- function(x) 'factor'
#' @export
vec_tween_class.character <- function(x) {
  colour <- try(suppressWarnings(decode_colour(x)), silent = TRUE)
  if (all(is.na(x)) || inherits(colour, 'try-error') || any(is.na(x) != is.na(colour[, 1])) || all(grepl('^(\\d|\\.)+$', x))) {
    'character'
  } else {
    'colour'
  }
}
#' @export
vec_tween_class.Date <- function(x) 'date'
#' @export
vec_tween_class.POSIXt <- function(x) 'datetime'
#' @export
vec_tween_class.list <- function(x) {
  if (all(vapply(x, is.numeric, logical(1)))) 'numlist'
  else 'list'
}

prepareTween <- function(data, n, ease) {
  if (!is.list(data)) {
    data <- as.list(data)
  }
  if (length(unique(lengths(data))) != 1) {
    stop('All elements in data must have the same length')
  }
  if (!all(ease %in% validEase)) {
    stop('ease must be the name of a valid easing function')
  }
  n <- c(rep(n, length.out = length(data) - 1) - 1, 1)
  ease <- c(rep(ease, length.out = length(data) - 1), 'constant')
  states <- data.frame(
    state = seq_along(data) - 1L,
    nframes = as.integer(n),
    ease = ease,
    stringsAsFactors = FALSE
  )
  list(
    data = data,
    states = states
  )
}

prepareTweenTranspose <- function(data, n, ease) {
  if (!is.list(data)) {
    data <- list(data)
  }
  if (!all(ease %in% validEase)) {
    stop('ease must be the name of a valid easing function')
  }
  n <- rep(n, length.out = length(data))
  n <- Map(function(n, l) {
    s <- floor(n / l)
    s <- rep(s, l)
    overhead <- n - sum(s)
    if (overhead) {
      s <- s + rep(floor(overhead / l), l)
      addInd <- seq_len(n - sum(s))
      s[addInd] <- s[addInd] + 1
    }
    c(s, 1)
  }, n = n - 1, l = lengths(data) - 1)
  n <- unlist(n)
  ease <- rep(ease, length.out = length(data))
  ease <- rep(ease, lengths(data) - 1)
  easeSplit <- split(ease, rep(seq_along(data), lengths(data) - 1))
  ease <- unlist(lapply(easeSplit, append, values = 'constant'))
  data <- as.list(unlist(data))
  states <- data.frame(
    state = seq_along(data) - 1L,
    nframes = as.integer(n),
    ease = ease,
    stringsAsFactors = FALSE
  )
  list(
    data = data,
    states = states
  )
}

first <- function(x) x[[1]]
`first<-` <- function(x, value) {
  x[[1]] <- value
  x
}
last <- function(x) x[[length(x)]]
`last<-` <- function(x, value) {
  x[[length(x)]] <- value
  x
}

#' Display an easing function
#'
#' This simple helper lets you explore how the different easing functions govern
#' the interpolation of data.
#'
#' @details
#' How transitions proceed between states are defined by an easing function. The
#' easing function converts the parameterized progression from one state to the
#' next to a new number between 0 and 1. `linear` easing is equivalent to
#' an identity function that returns the input unchanged. In addition there are
#' a range of additional easers available, each with three modifiers.
#'
#' \strong{Easing modifiers:}
#' \describe{
#'  \item{-in}{The easing function is applied as-is}
#'  \item{-out}{The easing function is applied in reverse}
#'  \item{-in-out}{The first half of the transition it is applied as-is, while
#'  in the last half it is reversed}
#' }
#'
#' \strong{Easing functions}
#' \describe{
#'  \item{quadratic}{Models a power-of-2 function}
#'  \item{cubic}{Models a power-of-3 function}
#'  \item{quartic}{Models a power-of-4 function}
#'  \item{quintic}{Models a power-of-5 function}
#'  \item{sine}{Models a sine function}
#'  \item{circular}{Models a pi/2 circle arc}
#'  \item{exponential}{Models an exponential function}
#'  \item{elastic}{Models an elastic release of energy}
#'  \item{back}{Models a pullback and relase}
#'  \item{bounce}{Models the bouncing of a ball}
#' }
#'
#' In addition to this function a good animated explanation can be found
#' [here](https://easings.net).
#'
#' @param ease The name of the easing function to display (see details)
#'
#' @return This function is called for its side effects
#'
#' @examples
#' # The default - identity
#' display_ease('linear')
#'
#' # A more fancy easer
#' display_ease('elastic-in')
#'
#' @importFrom graphics plot
#' @export
#'
display_ease <- function(ease) {
  easepoints <- tween_numeric(c(0, 1), 100, ease)[[1]]
  progress <- seq(0, 1, length.out = 100)
  plot(progress, easepoints, type = 'l', main = ease, xlab = 'In',
       ylab = 'Out', bty = 'n')
}

#' Extract a frame from a generator
#'
#' Using the generators in tweenr you can avoid calculating all needed frames up
#' front, which can be prohibitive in memory. With a generator you can use
#' `get_frame()` to extract any frame at a fractional location between 0 and 1
#' one by one as you need them. You can further get all raw data before and/or
#' after a given point in time using `get_raw_frames()`.
#'
#' @param generator A `frame_generator` object
#' @param at A scalar numeric between 0 and 1
#' @param before,after Scalar numerics that define the time before and after
#' `at` to search for raw data
#' @param ... Arguments passed on to methods
#'
#' @export
#'
#' @examples
#' data <- data.frame(
#'   x = c(1, 2, 2, 1, 2, 2),
#'   y = c(1, 2, 2, 2, 1, 1),
#'   time = c(1, 4, 8, 4, 8, 10),
#'   id = c(1, 1, 1, 2, 2, 2)
#' )
#'
#' gen <- gen_components(data, 'cubic-in-out', time = time, id = id)
#'
#' get_frame(gen, 0.3)
#'
#' get_raw_frames(gen, 0.5, before = 0.5, after = 0.2)
get_frame <- function(generator, at, ...) {
  UseMethod("get_frame")
}
#' @rdname get_frame
#' @export
get_raw_frames <- function(generator, at, before = 0, after = 0, ...) {
  UseMethod("get_raw_frames")
}

#' @details
#' tweenr is a small collection of functions to help you in creating
#' intermediary representations of your data, i.e. interpolating states of data.
#' As such it's a great match for packages such as animate and gganimate, since
#' it can work directly with data.frames of data, but it also provide fast and
#' efficient interpolaters for numeric, date, datetime and colour that are
#' vectorized and thus more efficient to use than the build in interpolation
#' functions (mainly [stats::approx()] and
#' [grDevices::colorRamp()]).
#'
#' The main functions for data.frames are [tween_states()],
#' [tween_elements()] and [tween_appear()], while the
#' standard interpolaters can be found at [tween()]
#'
#' @useDynLib tweenr, .registration = TRUE
'_PACKAGE'

#' Create frames based on individual element states
#'
#' This function creates tweens for each observation individually, in cases
#' where the data doesn't pass through collective states but consists of fully
#' independent transitions. Each observation is identified by an id and each
#' state must have a time associated with it.
#'
#' @param data A data.frame consisting at least of a column giving the
#' observation id, a column giving timepoints for each state and a column giving
#' the easing to apply when transitioning away from the state.
#'
#' @param time The name of the column holding timepoints
#'
#' @param group The name of the column holding the observation id
#'
#' @param ease The name of the column holding the easing function name
#'
#' @param timerange The range of time to span. If missing it will default to
#' \code{range(data[[time]])}
#'
#' @param nframes The number of frames to generate. If missing it will default
#' to `ceiling(diff(timerange) + 1)` (At least one frame for each
#' individual timepoint)
#'
#' @return A data.frame with the same columns as `data` except for the
#' group and ease columns, but replicated `nframes` times. Two additional
#' columns called `.frame` and `.group` will be added giving the frame
#' number and observation id for each row.
#'
#' @family data.frame tween
#'
#' @examples
#' data <- data.frame(
#'   x = c(1, 2, 2, 1, 2, 2),
#'   y = c(1, 2, 2, 2, 1, 1),
#'   time = c(1, 4, 10, 4, 8, 10),
#'   group = c(1, 1, 1, 2, 2, 2),
#'   ease = rep('cubic-in-out', 6)
#' )
#'
#' data <- tween_elements(data, 'time', 'group', 'ease', nframes = 100)
#'
#' @export
#'
tween_elements <- function(data, time, group, ease, timerange, nframes) {
  if (!all(data[[ease]] %in% validEase)) {
    stop("All names given in the easing column must be valid easers")
  }

  if (missing(timerange) || is.null(timerange)) {
    timerange <- range(data[[time]])
  }
  if (missing(nframes) || is.null(nframes)) {
    nframes <- ceiling(diff(timerange) + 1)
  }
  framelength <- diff(timerange) / nframes
  specialCols <- c(group, ease)
  data <- data[order(data[[group]], data[[time]]), ]
  groups <- unique(data[[group]])
  group <- match(data[[group]], groups)
  frame <- round((data[[time]] - timerange[1]) / framelength)
  ease <- as.character(data[[ease]])
  data <- data[, !names(data) %in% specialCols, drop = FALSE]

  colClasses <- col_classes(data)
  tweendata <- lapply(seq_along(data),  function(i) {
    d <- data[[i]]
    switch(
      colClasses[i],
      numeric = interpolate_numeric_element(d, group, frame, ease),
      logical = interpolate_logical_element(d, group, frame, ease),
      factor = interpolate_factor_element(d, group, frame, ease),
      character = interpolate_character_element(d, group, frame, ease),
      colour = interpolate_colour_element(d, group, frame, ease),
      date = interpolate_date_element(d, group, frame, ease),
      datetime = interpolate_datetime_element(d, group, frame, ease),
      constant = interpolate_constant_element(d, group, frame, ease),
      numlist = interpolate_numlist_element(d, group, frame, ease),
      list = interpolate_list_element(d, group, frame, ease)
    )
  })
  tweenInfo <- tweendata[[1]][, c('group', 'frame')]
  tweendata <- as.data.frame(lapply(tweendata, `[[`, i = 'data'))
  names(tweendata) <- names(data)
  tweendata$.frame <- tweenInfo$frame
  tweendata$.group <- groups[tweenInfo$group]
  attr(tweendata, 'framelength') <- framelength
  tweendata[order(tweendata$.frame, tweendata$.group), ]
}

interpolate_numeric_fill <- function(data, ease) {
  numeric_fill_interpolator(as.numeric(data), as.character(ease))
}

interpolate_logical_fill <- function(data, ease) {
  as.logical(round(numeric_fill_interpolator(data, ease)))
}

#' @importFrom farver decode_colour encode_colour
interpolate_colour_fill <- function(data, ease) {
  data <- decode_colour(data, alpha = TRUE, to = 'lab')
  data <- colour_fill_interpolator(data, as.character(ease))
  encode_colour(data[, 1:3, drop = FALSE], alpha = data[,4], from = 'lab')
}

interpolate_constant_fill <- function(data, ease) {
  constant_fill_interpolator(as.character(data), as.character(ease))
}

interpolate_character_fill <- interpolate_constant_fill

interpolate_date_fill <- function(data, ease) {
  as.Date(interpolate_numeric_fill(data, ease), origin = BASEDATE)
}

interpolate_datetime_fill <- function(data, ease) {
  if (inherits(data, 'POSIXlt')) {
    warning("POSIXlt converted to POSIXct")
    data <- as.POSIXct(data)
  }
  as.POSIXct(interpolate_numeric_fill(data, ease), origin = BASEDATETIME, tz = attr(data, 'tzone'))
}

interpolate_factor_fill <- function(data, ease) {
  all_levels <- levels(data)
  ord <- is.ordered(data)
  data <- interpolate_character_fill(data, ease)
  if (ord) ordered(data, all_levels) else factor(data, all_levels)
}

interpolate_list_fill <- function(data, ease) {
  new_data <- list_fill_interpolator(as.list(data), as.character(ease))
  attributes(new_data) <- attributes(data)
  new_data
}

interpolate_numlist_fill <- function(data, ease) {
  new_data <- numlist_fill_interpolator(lapply(data, as.numeric), as.character(ease))
  attributes(new_data) <- attributes(data)
  new_data
}

#' Compose tweening between states
#'
#' The `tween_state()` is a counterpart to `tween_states()` that is aimed at
#' letting you gradually build up a scene by composing state changes one by one.
#' This setup lets you take more control over each state change and allows you
#' to work with datasets with uneven number of rows, flexibly specifying what
#' should happen with entering and exiting data. `keep_state()` is a simple
#' helper for letting you pause at a state. `open_state()` is a shortcut from
#' tweening from an empty dataset with a given `enter()` function while
#' `close_state()` is the same but will instead tween into an empty dataset with
#' a given `exit()` function.
#'
#' @param .data A data.frame to start from. If `.data` is the result of a prior
#' tween, only the last frame will be used for the tween. The new tween will
#' then be added to the prior tween
#'
#' @param to A data.frame to end at. It must contain the same columns as .data
#' (exluding `.frame`)
#'
#' @param ease The easing function to use. Either a single string or one for
#' each column in the data set.
#'
#' @param nframes The number of frames to calculate for the tween
#'
#' @param id The column to match observations on. If `NULL` observations will be
#' matched by position. See the *Match, Enter, and Exit* section for more
#' information.
#'
#' @param enter,exit functions that calculate a start state for new observations
#' that appear in `to` or an end state for observations that are not present in
#' `to`. If `NULL` the new/old observations will not be part of the tween. The
#' function gets a data.frame with either the start state of the exiting
#' observations, or the end state of the entering observations and must return
#' a modified version of that data.frame. See the *Match, Enter, and Exit*
#' section for more information.
#'
#' @return A data.frame containing all the intermediary states in the tween,
#' each state will be enumerated by the `.frame` column
#'
#' @section Match, Enter, and Exit:
#' When there are discrepancies between the two states to tweeen between you
#' need a way to resolve the discrepancy before calculating the intermediary
#' states. With discrepancies we mean that some data points are present in the
#' start state and not in the end state, and/or some are present in the end
#' state but not in the start state. A simple example is that the start state
#' contains 100 rows and the end state contains 70. There are 30 missing rows
#' that we need to do something about before we can calculate the tween.
#'
#' **Making pairs**
#' The first question to answer is "How do we know which observations are
#' disappearing (*exiting*) and/or appearing (*entering*)?". This is done with
#' the `id` argument which should give a column name to match rows between the
#' two states on. If `id = NULL` the rows will be matched by position (in the
#' above example the last 30 rows in the start state will be entering). The `id`
#' column must only contain unique values in order to work.
#'
#' **Making up states**
#' Once the rows in each state has been paired you'll end up with three sets of
#' data. One containing rows that is present in both the start and end state,
#' one containing rows only present in the start state, and one only containing
#' rows present in the end state. The first group is easy - here you just tween
#' between each rows - but for the other two we'll need some state to start or
#' end the tween with. This is really the purpose of the `enter` and `exit`
#' functions. They take a data frame containing the subset of data that has not
#' been matched and must return a new data frame giving the state that these
#' rows must be tweened from/into. A simple example could be an `enter` function
#' that sets the variable giving the opacity in the plot to 0 - this will make
#' the new points fade into view during the transition.
#'
#' **Ignoring discrepancies**
#' The default values for `enter` and `exit` is `NULL`. This value indicate that
#' non-matching rows should simply be ignored for the transition and simply
#' appear in the last frame of the tween. This is the default.
#'
#' @importFrom rlang enquo
#' @importFrom vctrs vec_rbind vec_cbind
#' @export
#'
#' @examples
#' data1 <- data.frame(
#'   x = 1:20,
#'   y = 0,
#'   colour = 'forestgreen',
#'   stringsAsFactors = FALSE
#' )
#' data2 <- data1
#' data2$x <- 20:1
#' data2$y <- 1
#'
#' data <- data1 %>%
#'   tween_state(data2, 'linear', 50) %>%
#'   keep_state(20) %>%
#'   tween_state(data1, 'bounce-out', 50)
#'
#' # Using enter and exit (made up numbers)
#' df1 <- data.frame(
#'   country = c('Denmark', 'Sweden', 'Norway'),
#'   population = c(5e6, 10e6, 3.5e6)
#' )
#' df2 <- data.frame(
#'   country = c('Denmark', 'Sweden', 'Norway', 'Finland'),
#'   population = c(6e6, 10.5e6, 4e6, 3e6)
#' )
#' df3 <- data.frame(
#'   country = c('Denmark', 'Norway'),
#'   population = c(10e6, 6e6)
#' )
#' to_zero <- function(x) {
#'   x$population <- 0
#'   x
#' }
#' pop_devel <- df1 %>%
#'   tween_state(df2, 'cubic-in-out', 50, id = country, enter = to_zero) %>%
#'   tween_state(df3, 'cubic-in-out', 50, id = country, enter = to_zero,
#'               exit = to_zero)
#'
tween_state <- function(.data, to, ease, nframes, id = NULL, enter = NULL, exit = NULL) {
  .data[] <- lapply(.data, fix_old_mapped_discrete)
  to[] <- lapply(to, fix_old_mapped_discrete)
  from <- .get_last_frame(.data)
  from$.phase <- rep('raw', length.out = nrow(from))
  to$.phase <- rep('raw', length.out = nrow(to))
  to$.id <- rep(NA_integer_, length.out = nrow(to))
  id <- enquo(id)
  if (.has_frames(.data)) nframes <- nframes + 1
  if (!setequal(names(from), names(to))) {
    stop('from and to must have identical columns', call. = FALSE)
  }

  if (nrow(from) == 0 && nrow(to) == 0) {
    return(.with_prior_frames(.data, from, nframes))
  }

  to <- to[, match(names(from), names(to)), drop = FALSE]
  if (length(ease) == 1) ease <- rep(ease, ncol(from) - 2)
  if (length(ease) == ncol(from) - 2) {
    ease <- c(ease, 'linear', 'linear') # To account for .phase and .id columns
  } else {
    stop('Ease must be either a single string or one for each column', call. = FALSE)
  }
  if (!is_integerish(nframes, 1L)) {
    stop("`nframes` must be a single count", call. = FALSE)
  }

  classes <- if (nrow(from) == 0) col_classes(to) else col_classes(from)
  if (nrow(from) > 0 && nrow(to) > 0) {
    to_classes <- col_classes(to)
    mismatch <- to_classes != classes
    for (i in which(mismatch)) {
      all_na_to <- all(is.na(to[[i]]))
      all_na_from <- all(is.na(from[[i]]))
      if (all_na_from) {
        storage.mode(from[[i]]) <- storage.mode(to[[i]])
      } else if (all_na_to) {
        storage.mode(to[[i]]) <- storage.mode(from[[i]])
      } else {
        stop('The ', names(to)[i], 'column differs in type between the two inputs', call. = FALSE)
      }
    }
  }
  full_set <- .complete_states(from, to, id, enter, exit, .max_id(.data))

  to$.id <- full_set$orig_to

  if (nrow(full_set$from) != 0) {
    tweendata <- lapply(seq_along(classes), function(i) {
      d <- list(full_set$from[[i]], full_set$to[[i]])
      state <- simple_state(as.integer(nframes), ease[i])
      switch(
        classes[i],
        numeric = interpolate_numeric_state(d, state),
        logical = interpolate_logical_state(d, state),
        factor = interpolate_factor_state(d, state),
        character = interpolate_character_state(d, state),
        colour = interpolate_colour_state(d, state),
        date = interpolate_date_state(d, state),
        datetime = interpolate_datetime_state(d, state),
        constant = interpolate_constant_state(d, state),
        numlist = interpolate_numlist_state(d, state),
        list = interpolate_list_state(d, state),
        phase = get_phase_state(d, state)
      )
    })
    tweendata <- structure(tweendata, names = names(full_set$from), row.names = .set_row_names(length(tweendata[[1]])), class = 'data.frame')
    tweendata$.frame <- rep(seq_len(nframes - 1), each = nrow(full_set$from))
  } else {
    tweendata <- full_set$from
    tweendata$.frame <- integer(0)
  }
  from[classes == "constant"] <- lapply(from[classes == "constant"], as.character)
  to[classes == "constant"] <- lapply(to[classes == "constant"], as.character)
  tweendata <- vec_rbind(
    if (nframes > 1) vec_cbind(from, .frame = rep(1, nrow(from))) else NULL,
    tweendata[tweendata$.frame != 1, , drop = FALSE],
    vec_cbind(to, .frame = rep(nframes, nrow(to)))
  )
  .with_prior_frames(.data, tweendata, nframes)
}
#' @rdname tween_state
#' @export
keep_state <- function(.data, nframes) {
  state <- .get_last_frame(.data)
  state$.phase <- rep('raw', length.out = nrow(state))
  if (.has_frames(.data)) nframes <- nframes + 1
  if (nrow(state) == 0) {
    return(.with_prior_frames(.data, state, nframes))
  }
  states <- state[rep(seq_len(nrow(state)), nframes), , drop = FALSE]
  states$.phase[seq_len(nrow(state) * (nframes - 1))] <- 'static'
  states$.frame <- rep(seq_len(nframes), each = nrow(state))
  .with_prior_frames(.data, states, nframes)
}
#' @rdname tween_state
#' @export
open_state <- function(.data, ease, nframes, enter) {
  to <- .get_first_frame(.data)
  if (.has_frames(.data)) nframes <- nframes + 1
  tweendata <- tween_state(to[0, , drop = FALSE], to, ease, nframes, enter = enter)
  .with_later_frames(.data, tweendata, nframes)
}
#' @rdname tween_state
#' @export
close_state <- function(.data, ease, nframes, exit) {
  from <- .get_last_frame(.data)
  if (.has_frames(.data)) nframes <- nframes + 1
  tweendata <- tween_state(from, from[0, , drop = FALSE], ease, nframes, exit = exit)
  .with_prior_frames(.data, tweendata, nframes)
}
#' Helpers for working with tweened data
#'
#' These are internal helpers for extracting and inserting data into a
#' data.frame of tweened states.
#'
#' @param data,prior,later A data.frame. If a `.frame` column exists it will be interpreted
#' as a data.frame containing multiple states
#'
#' @param new_tween The result of a tweening
#'
#' @return A data.frame
#' @keywords internal
#' @export
#'
.get_last_frame <- function(data) {
  nframes <- attr(data, 'nframes')
  data <- if (!is.null(nframes)) {
    data[data$.frame == nframes, names(data) != '.frame', drop = FALSE]
  } else if ('.frame' %in% names(data)) {
    data[data$.frame == max(data$.frame), names(data) != '.frame', drop = FALSE]
  } else {
    data
  }
  if (is.null(data$.id)) {
    data$.id <- seq_len(nrow(data))
  }
  data
}
#' @rdname dot-get_last_frame
#' @export
.get_first_frame <- function(data) {
  data <- if ('.frame' %in% names(data)) {
    data[data$.frame == 1, names(data) != '.frame', drop = FALSE]
  } else {
    data
  }
  if (is.null(data$.id)) {
    data$.id <- seq_len(nrow(data))
  }
  data
}
#' @rdname dot-get_last_frame
#' @export
.with_prior_frames <- function(prior, new_tween, nframes) {
  nframes_before <- attr(prior, 'nframes')
  if (is.null(nframes_before) && nrow(prior) > 0 && '.frame' %in% names(prior)) nframes_before <- max(prior$.frame)
  frames <- if (!is.null(nframes_before)) {
    prior <- prior[prior$.frame != nframes_before, , drop = FALSE]
    new_tween$.frame <- new_tween$.frame + nframes_before - 1
    if (is.character(prior$.id)) new_tween$.id <- as.character(new_tween$.id)
    else if (is.character(new_tween$.id)) prior$.id <- as.character(prior$.id)
    vec_rbind(prior, new_tween)
  } else {
    nframes_before <- 1
    new_tween
  }
  attr(frames, 'nframes') <- nframes + nframes_before - 1
  attr(frames, 'max_id') <- find_max_id(prior, new_tween)
  frames
}
#' @rdname dot-get_last_frame
#' @export
.with_later_frames <- function(later, new_tween, nframes) {
  nframes_before <- attr(later, 'nframes')
  nframes_before <- if (is.null(nframes_before) && nrow(later) > 0 && '.frame' %in% names(later)) max(later$.frame) else 1
  frames <- if ('.frame' %in% names(later)) {
    later <- later[later$.frame != 1, , drop = FALSE]
    later$.frame <- later$.frame + max(new_tween$.frame)
    if (is.character(later$.id)) new_tween$.id <- as.character(new_tween$.id)
    else if (is.character(new_tween$.id)) later$.id <- as.character(later$.id)
    vec_rbind(new_tween, later)
  } else {
    new_tween
  }
  attr(frames, 'nframes') <- nframes + nframes_before - 1
  attr(frames, 'max_id') <- find_max_id(later, new_tween)
  frames
}
find_max_id <- function(data, new) {
  max_new <- if (nrow(new) == 0) 0 else max(new$.id)
  max(max_new, .max_id(data))
}
#' Get the highest id occuring in a dataset
#'
#' This is helper for `tween_state` related functions to get the currently
#' highest `.id` in a frame collection
#'
#' @param data A data.frame as returned by `tween_state`
#'
#' @return An integer giving the currently highest id
#'
#' @keywords internal
#' @export
.max_id <- function(data) {
  max_id <- attr(data, 'max_id')
  if (is.null(max_id) && nrow(data) > 0 && !is.null(data$.id)) max_id <- max(data$.id)
  else max_id <- nrow(data)
  max_id
}
#' Fill in missing rows using enter and exit functions
#'
#' This function figures out which rows are missing in either state and applies
#' the provided `enter` and `exit`  functions to fill in the blanks and provide
#' a 1-to-1 relation between the rows in `from` and `to`.
#'
#' @param from,to Data.frames to tween between
#'
#' @param id The name of the column that holds the matching id
#'
#' @param enter,exit functions to fill out missing rows in `from` and `to`
#' respectively
#'
#' @return A list with the elements `from` and `to` holding the filled out
#' versions of `from` and `to`
#'
#' @keywords internal
#' @importFrom rlang eval_tidy %||% as_function
#' @export
.complete_states <- function(from, to, id, enter, exit, max_id) {
  from_id <- eval_tidy(id, from) %||% seq_len(nrow(from))
  to_id <- eval_tidy(id, to) %||% seq_len(nrow(to))
  if (length(from_id) != nrow(from) || length(to_id) != nrow(to)) {
    stop('id must match the length of the data', call. = FALSE)
  }
  n_to <- nrow(to)
  if (anyDuplicated(from_id) || anyDuplicated(to_id) || !setequal(from_id, to_id)) {
    from_id <- paste(from_id, count_occourance(from_id), sep = '_')
    to_id <- paste(to_id, count_occourance(to_id), sep = '_')
    entering <- !to_id %in% from_id
    exiting <- !from_id %in% to_id

    if (is.null(enter) || sum(entering) == 0) {
      to <- to[!entering, , drop = FALSE]
      to_id <- to_id[!entering]
      enters <- to[0, , drop = FALSE]
      enter_id <- to_id[0]
    } else {
      enters <- as_function(enter)(to[entering, , drop = FALSE])
      enters$.phase <- 'enter'
      enter_id <- to_id[entering]
    }
    if (is.null(exit) || sum(exiting) == 0) {
      from <- from[!exiting, , drop = FALSE]
      from_id <- from_id[!exiting]
      exits <- from[0, , drop = FALSE]
      exit_id <- from_id[0]
    } else {
      exits <- as_function(exit)(from[exiting, , drop = FALSE])
      exits$.phase <- 'exit'
      exit_id <- from_id[exiting]
    }
    from <- vec_rbind(
      safe_df(from),
      safe_df(enters)
    )
    from_id <- c(from_id, enter_id)
    to <- vec_rbind(
      safe_df(to),
      safe_df(exits)
    )
    to_id <- c(to_id, exit_id)
  }
  from$.id[is.na(from$.id)] <- seq_len(sum(is.na(from$.id))) + max_id
  orig_to_id <- from$.id[match(to_id, from_id)][seq_len(n_to)]
  to <- to[match(from_id, to_id), , drop = FALSE]
  to$.id <- from$.id

  list(from = from, to = to, orig_to = orig_to_id)
}
#' @rdname dot-get_last_frame
#' @export
.has_frames <- function(data) {
  !is.null(attr(data, 'nframes')) || !is.null(data$.frame)
}
simple_state <- function(n, ease) {
  data.frame(state = c(0L, 1L), nframes = c(n - 1L, 0L), ease = c(ease, 'constant'), stringsAsFactors = FALSE)
}

count_occourance <- function(x) {
  if (length(x) == 0) return(integer(0))
  unsplit(lapply(split(x, x), seq_along), x)
}

fix_old_mapped_discrete <- function(x) {
  if (inherits(x, 'mapped_discrete') && storage.mode(x) == 'integer') {
    storage.mode(x) <- 'double'
  }
  x
}

safe_df <- function(x) {
  if (nrow(x) > 0) return(x)
  x[] <- lapply(x, function(col) if (is.logical(col)) vctrs::unspecified() else col)
  x
}

context("state")

df1 <- data.frame(x = 1:2, y = 4:5, col = 'black', type = letters[1:2], stringsAsFactors = FALSE)
df2 <- data.frame(x = 11:12, y = 14:15, col = 'white', type = letters[1], stringsAsFactors = FALSE)

test_that("tween_state works", {
  tween <- tween_state(df1, df2, ease = 'linear', nframes = 5)
  expect_equal(max(tween$.frame), 5)
  expect_true(all(tween$.phase[c(1:2, 9:10)] == 'raw'))
  expect_true(all(tween$.phase[c(3:8)] == 'transition'))
  expect_true(all(tween$.id == rep(1:2, 5)))
  expect_equal(tween$col[5], '#777777')
  expect_equal(tween$x[7], 8.5)
  expect_equal(tween$type[4:5], c('b', 'a'))
})

test_that("keep_state works", {
  expect_warning(
    keep <- keep_state(df1, 5),
    NA
  )
  expect_equal(max(keep$.frame), 5)
  expect_true(all(keep$.phase[c(9:10)] == 'raw'))
  expect_true(all(keep$.phase[c(1:8)] == 'static'))
})

test_that("enter/exit works", {
  tween <- tween_state(df1, df2[1,, drop = FALSE], 'linear', 5, exit = function(df) {
    df$x <- 0
    df$col <- 'red'
    df
  })
  expect_equal(nrow(tween), 9)
  expect_true(all(tween$.phase[c(4,6,8)] == 'exit'))
  expect_equal(tween$col[8], '#BB1909')
  expect_equal(tween$x[8], 0.5)
})

interpolate_numeric_state <- function(data, states) {
  res <- numeric_state_interpolator(lapply(data, as.numeric), states)
  c(data[[1]][0], res)
}

interpolate_logical_state <- function(data, states) {
  res <- numeric_state_interpolator(lapply(data, as.numeric), states)
  as.logical(round(res))
}

#' @importFrom farver decode_colour encode_colour
interpolate_colour_state <- function(data, states) {
  data <- lapply(data, decode_colour, alpha = TRUE, to = 'lab')
  data <- colour_state_interpolator(data, states)
  encode_colour(data[, 1:3, drop = FALSE], alpha = data[,4], from = 'lab')
}

interpolate_constant_state <- function(data, states) {
  constant_state_interpolator(lapply(data, as.character), states)
}
interpolate_character_state <- interpolate_constant_state

interpolate_date_state <- function(data, states) {
  as.Date(interpolate_numeric_state(data, states), origin = BASEDATE)
}

interpolate_datetime_state <- function(data, states) {
  if (inherits(data[[1]], 'POSIXlt')) {
    warning("POSIXlt converted to POSIXct")
    data[[1]] <- as.POSIXct(data[[1]])
  }
  tz <- attr(data[[1]], 'tzone')
  as.POSIXct(interpolate_numeric_state(data, states), origin = BASEDATETIME, tz = tz)
}

interpolate_factor_state <- function(data, states) {
  all_levels <- Reduce(union, lapply(data, levels))
  ord <- is.ordered(data[[1]])
  data <- interpolate_character_state(data, states)
  if (ord) ordered(data, all_levels) else factor(data, all_levels)
}

interpolate_list_state <- function(data, states) {
  new_data <- list_state_interpolator(data, states)
  attributes(new_data) <- attributes(data)
  new_data
}

interpolate_numlist_state <- function(data, states) {
  new_data <- numlist_state_interpolator(lapply(data, lapply, as.numeric), states)
  attributes(new_data) <- attributes(data)
  new_data
}

get_phase_state <- function(data, states) {
  phase_state_interpolator(lapply(data, as.character), states)
}

#' Generator for interpolating between two data frames
#'
#' This is a generator version of [tween_at()] with the additional functionality
#' of supporting enter and exit functions. It returns a generator that can be
#' used with [get_frame()] and [get_raw_frames()] to extract frames for a
#' specific time point scaled between 0 and 1.
#'
#' @inheritParams tween_at
#' @inheritParams tween_state
#'
#' @return A `keyframe_generator` object
#'
#' @family Other generators
#'
#' @export
#'
#' @examples
#' gen <- gen_at(mtcars[1:6, ], mtcars[6:1, ], 'cubic-in-out')
#'
#' get_frame(gen, 0.3)
gen_at <- function(from, to, ease, id = NULL, enter = NULL, exit = NULL) {
  gen <- gen_keyframe(from, 0)
  add_keyframe(gen, to, ease = ease, length = 1, id = {{ id }}, enter = enter, exit = exit)
}

#' Generator for tweening along a variable
#'
#' This is a generator version of [tween_along()]. It returns a generator that
#' can be used with [get_frame()] and [get_raw_frames()] to extract frames for
#' a specific time point scaled between 0 and 1.
#'
#' @inheritParams tween_along
#'
#' @return An `along_generator` object
#'
#' @family Other generators
#'
#' @export
#' @importFrom rlang eval_tidy enquo
#'
#' @examples
#' # Default behaviour
#' gen <- gen_along(airquality, ease = "linear", along = Day, id = Month)
#' get_frame(gen, 0.22)
#'
#' # Overwrite keep_last or history in get_frame
#' get_frame(gen, 0.67, history = FALSE)
gen_along <- function(.data, ease, along, id = NULL, range = NULL, history = TRUE, keep_last = FALSE) {
  along <- enquo(along)
  along <- as.numeric(eval_tidy(along, .data))
  id <- enquo(id)
  id <- if (quo_is_null(id)) rep(1, nrow(.data)) else eval_tidy(id, .data)

  .data$.phase <- NULL
  if (length(ease) == 1) ease <- rep(ease, ncol(.data))
  if (length(ease) == ncol(.data)) {
    ease <- c(ease, 'linear') # To account for .phase column
  } else {
    stop('Ease must be either a single string or one for each column', call. = FALSE)
  }

  .data$.phase <- rep_len(factor("raw", levels = PHASE_LEVELS), nrow(.data))
  d_order <- order(id, along)
  if (is.null(range)) range <- range(along)
  class(.data) <- c(c("along_generator", "frame_generator"), class(.data))
  generator_settings(.data) <- list(
    data = .data[d_order, ],
    id = id[d_order],
    along = along[d_order],
    range = range,
    ease_type = ease,
    history = history,
    keep_last = keep_last,
    col_types = col_classes(.data)
  )
  .data
}
#' @export
get_frame.along_generator <- function(generator, at, ..., history = NULL, keep_last = NULL) {
  d <- generator_settings(generator)
  # clamp between 0 and 1
  at <- min(max(at, 0), 1)
  range <- d$range

  # normalise to range
  at <- (range[2] - range[1]) * at + range[1]

  data <- gen_data(generator)
  ease <- ease_type(generator)
  type <- col_types(generator)
  if (is.null(history)) history <- d$history
  if (is.null(keep_last)) keep_last <- d$keep_last

  frame <- lapply(seq_along(data),  function(i) {
    data <- data[[i]]
    e <- ease[i]
    switch(
      type[i],
      numeric = interpolate_numeric_along(data, d$id, d$along, at, e, history, keep_last)$data,
      logical = interpolate_logical_along(data, d$id, d$along, at, e, history, keep_last)$data,
      factor = interpolate_factor_along(data, d$id, d$along, at, e, history, keep_last)$data,
      character = interpolate_character_along(data, d$id, d$along, at, e, history, keep_last)$data,
      colour = interpolate_colour_along(data, d$id, d$along, at, e, history, keep_last)$data,
      date = interpolate_date_along(data, d$id, d$along, at, e, history, keep_last)$data,
      datetime = interpolate_datetime_along(data, d$id, d$along, at, e, history, keep_last)$data,
      constant = interpolate_constant_along(data, d$id, d$along, at, e, history, keep_last)$data,
      numlist = interpolate_numlist_along(data, d$id, d$along, at, e, history, keep_last)$data,
      list = interpolate_list_along(data, d$id, d$along, at, e, history, keep_last)$data,
      phase = get_phase_along(d$id, d$along, at, history, keep_last)$data
    )
  })
  structure(frame, names = names(data), row.names = .set_row_names(length(frame[[1]])), class = 'data.frame')
}
#' @export
get_raw_frames.along_generator <- function(generator, at, before = 0, after = 0, ...) {
  d <- generator_settings(generator)

  # clamp between 0 and 1
  before <- min(max(at - before, 0), 1)
  after <- min(max(at + after, 0), 1)
  at <- min(max(at, 0), 1)
  range <- d$range

  # normalise to generator time
  at <- (range[2] - range[1]) * at + range[1]
  before <- (range[2] - range[1]) * before + range[1]
  after <- (range[2] - range[1]) * after + range[1]

  # Find before and after
  before <- d$along >= before & d$along < at
  after <- d$along > at & d$along <= after

  list(
    before = d$data[before, , drop = FALSE],
    after = d$data[after, , drop = FALSE]
  )
}

#' @export
convert_generator.along_generator <- function(x) {
  data <- gen_data(x)
  settings <- list(
    attributes = attributes(x),
    data = data.frame(x),
    converter = function(x, settings) {
      data <- settings$data
      attributes(data) <- settings$attributes
      gen_data(data) <- data.frame(x)
      data
    }
  )
  list(data = data, settings = settings)
}

# Generated by cpp11: do not edit by hand

numeric_along_interpolator <- function(data, group, time, history, keep_last, frames, ease) {
  .Call(`_tweenr_numeric_along_interpolator`, data, group, time, history, keep_last, frames, ease)
}

colour_along_interpolator <- function(data, group, time, history, keep_last, frames, ease) {
  .Call(`_tweenr_colour_along_interpolator`, data, group, time, history, keep_last, frames, ease)
}

constant_along_interpolator <- function(data, group, time, history, keep_last, frames, ease) {
  .Call(`_tweenr_constant_along_interpolator`, data, group, time, history, keep_last, frames, ease)
}

list_along_interpolator <- function(data, group, time, history, keep_last, frames, ease) {
  .Call(`_tweenr_list_along_interpolator`, data, group, time, history, keep_last, frames, ease)
}

numlist_along_interpolator <- function(data, group, time, history, keep_last, frames, ease) {
  .Call(`_tweenr_numlist_along_interpolator`, data, group, time, history, keep_last, frames, ease)
}

phase_along_interpolator <- function(group, time, history, keep_last, frames) {
  .Call(`_tweenr_phase_along_interpolator`, group, time, history, keep_last, frames)
}

numeric_at_interpolator <- function(from, to, at, ease) {
  .Call(`_tweenr_numeric_at_interpolator`, from, to, at, ease)
}

colour_at_interpolator <- function(from, to, at, ease) {
  .Call(`_tweenr_colour_at_interpolator`, from, to, at, ease)
}

constant_at_interpolator <- function(from, to, at, ease) {
  .Call(`_tweenr_constant_at_interpolator`, from, to, at, ease)
}

list_at_interpolator <- function(from, to, at, ease) {
  .Call(`_tweenr_list_at_interpolator`, from, to, at, ease)
}

numlist_at_interpolator <- function(from, to, at, ease) {
  .Call(`_tweenr_numlist_at_interpolator`, from, to, at, ease)
}

numeric_at_t_interpolator <- function(from, to, at, ease) {
  .Call(`_tweenr_numeric_at_t_interpolator`, from, to, at, ease)
}

colour_at_t_interpolator <- function(from, to, at, ease) {
  .Call(`_tweenr_colour_at_t_interpolator`, from, to, at, ease)
}

constant_at_t_interpolator <- function(from, to, at, ease) {
  .Call(`_tweenr_constant_at_t_interpolator`, from, to, at, ease)
}

list_at_t_interpolator <- function(from, to, at, ease) {
  .Call(`_tweenr_list_at_t_interpolator`, from, to, at, ease)
}

numlist_at_t_interpolator <- function(from, to, at, ease) {
  .Call(`_tweenr_numlist_at_t_interpolator`, from, to, at, ease)
}

numeric_element_interpolator <- function(data, group, frame, ease) {
  .Call(`_tweenr_numeric_element_interpolator`, data, group, frame, ease)
}

colour_element_interpolator <- function(data, group, frame, ease) {
  .Call(`_tweenr_colour_element_interpolator`, data, group, frame, ease)
}

constant_element_interpolator <- function(data, group, frame, ease) {
  .Call(`_tweenr_constant_element_interpolator`, data, group, frame, ease)
}

list_element_interpolator <- function(data, group, frame, ease) {
  .Call(`_tweenr_list_element_interpolator`, data, group, frame, ease)
}

numlist_element_interpolator <- function(data, group, frame, ease) {
  .Call(`_tweenr_numlist_element_interpolator`, data, group, frame, ease)
}

phase_element_interpolator <- function(data, group, frame, ease) {
  .Call(`_tweenr_phase_element_interpolator`, data, group, frame, ease)
}

numeric_element_at_interpolator <- function(data, group, time, at, ease) {
  .Call(`_tweenr_numeric_element_at_interpolator`, data, group, time, at, ease)
}

colour_element_at_interpolator <- function(data, group, time, at, ease) {
  .Call(`_tweenr_colour_element_at_interpolator`, data, group, time, at, ease)
}

constant_element_at_interpolator <- function(data, group, time, at, ease) {
  .Call(`_tweenr_constant_element_at_interpolator`, data, group, time, at, ease)
}

list_element_at_interpolator <- function(data, group, time, at, ease) {
  .Call(`_tweenr_list_element_at_interpolator`, data, group, time, at, ease)
}

numlist_element_at_interpolator <- function(data, group, time, at, ease) {
  .Call(`_tweenr_numlist_element_at_interpolator`, data, group, time, at, ease)
}

phase_element_at_interpolator <- function(data, group, time, at, ease) {
  .Call(`_tweenr_phase_element_at_interpolator`, data, group, time, at, ease)
}

numeric_fill_interpolator <- function(data, ease) {
  .Call(`_tweenr_numeric_fill_interpolator`, data, ease)
}

colour_fill_interpolator <- function(data, ease) {
  .Call(`_tweenr_colour_fill_interpolator`, data, ease)
}

constant_fill_interpolator <- function(data, ease) {
  .Call(`_tweenr_constant_fill_interpolator`, data, ease)
}

list_fill_interpolator <- function(data, ease) {
  .Call(`_tweenr_list_fill_interpolator`, data, ease)
}

numlist_fill_interpolator <- function(data, ease) {
  .Call(`_tweenr_numlist_fill_interpolator`, data, ease)
}

numeric_state_interpolator <- function(data, states) {
  .Call(`_tweenr_numeric_state_interpolator`, data, states)
}

colour_state_interpolator <- function(data, states) {
  .Call(`_tweenr_colour_state_interpolator`, data, states)
}

constant_state_interpolator <- function(data, states) {
  .Call(`_tweenr_constant_state_interpolator`, data, states)
}

list_state_interpolator <- function(data, states) {
  .Call(`_tweenr_list_state_interpolator`, data, states)
}

numlist_state_interpolator <- function(data, states) {
  .Call(`_tweenr_numlist_state_interpolator`, data, states)
}

phase_state_interpolator <- function(data, states) {
  .Call(`_tweenr_phase_state_interpolator`, data, states)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_state.R
\name{.complete_states}
\alias{.complete_states}
\title{Fill in missing rows using enter and exit functions}
\usage{
.complete_states(from, to, id, enter, exit, max_id)
}
\arguments{
\item{from, to}{Data.frames to tween between}

\item{id}{The name of the column that holds the matching id}

\item{enter, exit}{functions to fill out missing rows in \code{from} and \code{to}
respectively}
}
\value{
A list with the elements \code{from} and \code{to} holding the filled out
versions of \code{from} and \code{to}
}
\description{
This function figures out which rows are missing in either state and applies
the provided \code{enter} and \code{exit}  functions to fill in the blanks and provide
a 1-to-1 relation between the rows in \code{from} and \code{to}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gen_components.R
\name{gen_components}
\alias{gen_components}
\title{Generator for tweening components separately from each other}
\usage{
gen_components(
  .data,
  ease,
  nframes,
  time,
  id = NULL,
  range = NULL,
  enter = NULL,
  exit = NULL,
  enter_length = 0,
  exit_length = 0
)
}
\arguments{
\item{.data}{A data.frame with components at different stages}

\item{ease}{The easing function to use. Either a single string or one for
each column in the data set.}

\item{nframes}{The number of frames to calculate for the tween}

\item{time}{An unquoted expression giving the timepoint for the different
stages of the components. Will be evaluated in the context of \code{.data} so can
refer to a column from that}

\item{id}{An unquoted expression giving the component id for each row. Will
be evaluated in the context of \code{.data} so can refer to a column from that}

\item{range}{The range of time points to include in the tween. If \code{NULL} it
will use the range of \code{time}}

\item{enter, exit}{functions that calculate a start state for new observations
that appear in \code{to} or an end state for observations that are not present in
\code{to}. If \code{NULL} the new/old observations will not be part of the tween. The
function gets a data.frame with either the start state of the exiting
observations, or the end state of the entering observations and must return
a modified version of that data.frame. See the \emph{Match, Enter, and Exit}
section for more information.}

\item{enter_length, exit_length}{The lenght of the opening and closing
transitions if \code{enter} and/or \code{exit} is given. Measured in the same units as
\code{time}}
}
\value{
A \code{component_generator} object
}
\description{
This is a generator versions of \code{\link[=tween_components]{tween_components()}}. It returns a generator
that can be used with \code{\link[=get_frame]{get_frame()}} and \code{\link[=get_raw_frames]{get_raw_frames()}} to extract frames
for a specific time point scaled between 0 and 1.
}
\examples{
from_zero <- function(x) {x$x <- 0; x}

data <- data.frame(
  x = c(1, 2, 2, 1, 2, 2),
  y = c(1, 2, 2, 2, 1, 1),
  time = c(1, 4, 8, 4, 8, 10),
  id = c(1, 1, 1, 2, 2, 2)
)

gen <- gen_components(data, 'cubic-in-out', time = time, id = id,
                      enter = from_zero, enter_length = 4)

get_frame(gen, 0.3)
}
\seealso{
Other Other generators: 
\code{\link{gen_along}()},
\code{\link{gen_at}()},
\code{\link{gen_events}()},
\code{\link{gen_keyframe}()}
}
\concept{Other generators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_state.R
\name{.max_id}
\alias{.max_id}
\title{Get the highest id occuring in a dataset}
\usage{
.max_id(data)
}
\arguments{
\item{data}{A data.frame as returned by \code{tween_state}}
}
\value{
An integer giving the currently highest id
}
\description{
This is helper for \code{tween_state} related functions to get the currently
highest \code{.id} in a frame collection
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gen_at.R
\name{gen_at}
\alias{gen_at}
\title{Generator for interpolating between two data frames}
\usage{
gen_at(from, to, ease, id = NULL, enter = NULL, exit = NULL)
}
\arguments{
\item{from, to}{A data.frame or vector of the same type. If either is of
length/nrow 1 it will get repeated to match the length of the other}

\item{ease}{A character vector giving valid easing functions. Recycled to
match the ncol of \code{from}}

\item{id}{The column to match observations on. If \code{NULL} observations will be
matched by position. See the \emph{Match, Enter, and Exit} section for more
information.}

\item{enter, exit}{functions that calculate a start state for new observations
that appear in \code{to} or an end state for observations that are not present in
\code{to}. If \code{NULL} the new/old observations will not be part of the tween. The
function gets a data.frame with either the start state of the exiting
observations, or the end state of the entering observations and must return
a modified version of that data.frame. See the \emph{Match, Enter, and Exit}
section for more information.}
}
\value{
A \code{keyframe_generator} object
}
\description{
This is a generator version of \code{\link[=tween_at]{tween_at()}} with the additional functionality
of supporting enter and exit functions. It returns a generator that can be
used with \code{\link[=get_frame]{get_frame()}} and \code{\link[=get_raw_frames]{get_raw_frames()}} to extract frames for a
specific time point scaled between 0 and 1.
}
\examples{
gen <- gen_at(mtcars[1:6, ], mtcars[6:1, ], 'cubic-in-out')

get_frame(gen, 0.3)
}
\seealso{
Other Other generators: 
\code{\link{gen_along}()},
\code{\link{gen_components}()},
\code{\link{gen_events}()},
\code{\link{gen_keyframe}()}
}
\concept{Other generators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_at.R
\name{tween_at_t}
\alias{tween_at_t}
\title{Get several specific position between two states}
\usage{
tween_at_t(from, to, at, ease)
}
\arguments{
\item{from, to}{A data.frame or vector of the same type. If either is of
length/nrow 1 it will get repeated to match the length of the other}

\item{at}{A numeric vector with values between 0 and 1.}

\item{ease}{A character vector giving valid easing functions. Recycled to
match the ncol of \code{from}}
}
\value{
If \code{from}/\code{to} is a data.frame then a data.frame with the same
columns. If \code{from}/\code{to} is a vector then a vector.
}
\description{
This tween is a variation of \code{\link[=tween_at]{tween_at()}}. Instead of having \code{at} refer to
the tweening position of each row, each \code{at} will interpolate the full data
at that position.
}
\examples{
tween_at_t(mtcars[1:6, ], mtcars[6:1, ], runif(3), 'cubic-in-out')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_appear.R
\name{tween_appear}
\alias{tween_appear}
\title{Tween a data.frame of appearances}
\usage{
tween_appear(data, time, timerange, nframes)
}
\arguments{
\item{data}{A data.frame to tween}

\item{time}{The name of the column that holds the time dimension. This does
not need to hold time data in the strictest sence - any numerical type will
do}

\item{timerange}{The range of time to create the tween for. If missing it
will defaults to the range of the time column}

\item{nframes}{The number of frames to create for the tween. If missing it
will create a frame for each full unit in \code{timerange} (e.g.
\code{timerange = c(1, 10)} will give \code{nframes = 10})}
}
\value{
A data.frame as \code{data} but repeated \code{nframes} times and
with the additional columns \code{.age} and \code{.frame}
}
\description{
This function is intended for use when you have a data.frame of events at
different time points. This could be the appearance of an observation for
example. This function replicates your data \code{nframes} times and
calculates the duration of each frame. At each frame each row is
assigned an age based on the progression of frames and the entry point of in
time for that row. A negative age means that the row has not appeared yet.
}
\examples{
data <- data.frame(
  x = rnorm(100),
  y = rnorm(100),
  time = sample(50, 100, replace = TRUE)
)

data <- tween_appear(data, 'time', nframes = 200)

}
\seealso{
Other data.frame tween: 
\code{\link{tween_along}()},
\code{\link{tween_components}()},
\code{\link{tween_elements}()},
\code{\link{tween_events}()},
\code{\link{tween_states}()}
}
\concept{data.frame tween}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gen_along.R
\name{gen_along}
\alias{gen_along}
\title{Generator for tweening along a variable}
\usage{
gen_along(
  .data,
  ease,
  along,
  id = NULL,
  range = NULL,
  history = TRUE,
  keep_last = FALSE
)
}
\arguments{
\item{.data}{A data.frame with components at different stages}

\item{ease}{The easing function to use. Either a single string or one for
each column in the data set.}

\item{along}{The "time" point for each row}

\item{id}{An unquoted expression giving the component id for each row. Will
be evaluated in the context of \code{.data} so can refer to a column from that}

\item{range}{The range of time points to include in the tween. If \code{NULL} it
will use the range of \code{time}}

\item{history}{Should earlier datapoints be kept in subsequent frames}

\item{keep_last}{Should the last point of each id be kept beyond its time}
}
\value{
An \code{along_generator} object
}
\description{
This is a generator version of \code{\link[=tween_along]{tween_along()}}. It returns a generator that
can be used with \code{\link[=get_frame]{get_frame()}} and \code{\link[=get_raw_frames]{get_raw_frames()}} to extract frames for
a specific time point scaled between 0 and 1.
}
\examples{
# Default behaviour
gen <- gen_along(airquality, ease = "linear", along = Day, id = Month)
get_frame(gen, 0.22)

# Overwrite keep_last or history in get_frame
get_frame(gen, 0.67, history = FALSE)
}
\seealso{
Other Other generators: 
\code{\link{gen_at}()},
\code{\link{gen_components}()},
\code{\link{gen_events}()},
\code{\link{gen_keyframe}()}
}
\concept{Other generators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_frame.R
\name{get_frame}
\alias{get_frame}
\alias{get_raw_frames}
\title{Extract a frame from a generator}
\usage{
get_frame(generator, at, ...)

get_raw_frames(generator, at, before = 0, after = 0, ...)
}
\arguments{
\item{generator}{A \code{frame_generator} object}

\item{at}{A scalar numeric between 0 and 1}

\item{...}{Arguments passed on to methods}

\item{before, after}{Scalar numerics that define the time before and after
\code{at} to search for raw data}
}
\description{
Using the generators in tweenr you can avoid calculating all needed frames up
front, which can be prohibitive in memory. With a generator you can use
\code{get_frame()} to extract any frame at a fractional location between 0 and 1
one by one as you need them. You can further get all raw data before and/or
after a given point in time using \code{get_raw_frames()}.
}
\examples{
data <- data.frame(
  x = c(1, 2, 2, 1, 2, 2),
  y = c(1, 2, 2, 2, 1, 1),
  time = c(1, 4, 8, 4, 8, 10),
  id = c(1, 1, 1, 2, 2, 2)
)

gen <- gen_components(data, 'cubic-in-out', time = time, id = id)

get_frame(gen, 0.3)

get_raw_frames(gen, 0.5, before = 0.5, after = 0.2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_events.R
\name{tween_events}
\alias{tween_events}
\title{Transition in and out of events}
\usage{
tween_events(
  .data,
  ease,
  nframes,
  start,
  end = NULL,
  range = NULL,
  enter = NULL,
  exit = NULL,
  enter_length = 0,
  exit_length = 0
)
}
\arguments{
\item{.data}{A data.frame with components at different stages}

\item{ease}{The easing function to use. Either a single string or one for
each column in the data set.}

\item{nframes}{The number of frames to calculate for the tween}

\item{start, end}{The start (and potential end) of the event encoded in the
row, as unquoted expressions. Will be evaluated in the context of \code{.data} so
can refer to columns in it. If \code{end = NULL} the event will be without extend
and only visible in a single frame, unless \code{enter} and/or \code{exit} is given.}

\item{range}{The range of time points to include in the tween. If \code{NULL} it
will use the range of \code{time}}

\item{enter, exit}{functions that calculate a start state for new observations
that appear in \code{to} or an end state for observations that are not present in
\code{to}. If \code{NULL} the new/old observations will not be part of the tween. The
function gets a data.frame with either the start state of the exiting
observations, or the end state of the entering observations and must return
a modified version of that data.frame. See the \emph{Match, Enter, and Exit}
section for more information.}

\item{enter_length, exit_length}{The lenght of the opening and closing
transitions if \code{enter} and/or \code{exit} is given. Measured in the same units as
\code{time}}
}
\value{
A data.frame with the same columns as \code{.data} along with \code{.id} giving
the component id, \code{.phase} giving the state of each component in each frame,
and \code{.frame} giving the frame membership of each row.
}
\description{
This tweening function is a more powerful version of \code{\link[=tween_appear]{tween_appear()}}, with
support for newer features such as enter/exits and tween phase
identification. The tweener treats each row in the data as unique events in
time, and creates frames with the correct events present at any given time.
}
\examples{
d <- data.frame(
  x = runif(20),
  y = runif(20),
  time = runif(20),
  duration = runif(20, max = 0.1)
)
from_left <- function(x) {
  x$x <- -0.5
  x
}
to_right <- function(x) {
  x$x <- 1.5
  x
}

tween_events(d, 'cubic-in-out', 50, start = time, end = time + duration,
             enter = from_left, exit = to_right, enter_length = 0.1,
             exit_length = 0.05)

}
\seealso{
Other data.frame tween: 
\code{\link{tween_along}()},
\code{\link{tween_appear}()},
\code{\link{tween_components}()},
\code{\link{tween_elements}()},
\code{\link{tween_states}()}
}
\concept{data.frame tween}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gen_keyframe.R
\name{prepare_keyframes}
\alias{prepare_keyframes}
\title{Prepare keyframes for generator}
\usage{
prepare_keyframes(.data, keyframe)
}
\arguments{
\item{.data}{A keyframe generator}

\item{keyframe}{A keyframe to add}
}
\value{
A valid keyframe generator
}
\description{
Prepare keyframes for generator
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gen_events.R
\name{gen_events}
\alias{gen_events}
\title{Generator for tweening the appearance of elements}
\usage{
gen_events(
  .data,
  ease,
  start,
  end = NULL,
  range = NULL,
  enter = NULL,
  exit = NULL,
  enter_length = 0,
  exit_length = 0
)
}
\arguments{
\item{.data}{A data.frame with components at different stages}

\item{ease}{The easing function to use. Either a single string or one for
each column in the data set.}

\item{start, end}{The start (and potential end) of the event encoded in the
row, as unquoted expressions. Will be evaluated in the context of \code{.data} so
can refer to columns in it. If \code{end = NULL} the event will be without extend
and only visible in a single frame, unless \code{enter} and/or \code{exit} is given.}

\item{range}{The range of time points to include in the tween. If \code{NULL} it
will use the range of \code{time}}

\item{enter, exit}{functions that calculate a start state for new observations
that appear in \code{to} or an end state for observations that are not present in
\code{to}. If \code{NULL} the new/old observations will not be part of the tween. The
function gets a data.frame with either the start state of the exiting
observations, or the end state of the entering observations and must return
a modified version of that data.frame. See the \emph{Match, Enter, and Exit}
section for more information.}

\item{enter_length, exit_length}{The lenght of the opening and closing
transitions if \code{enter} and/or \code{exit} is given. Measured in the same units as
\code{time}}
}
\value{
A \code{component_generator} object
}
\description{
This is a generator version of \code{\link[=tween_events]{tween_events()}}. It returns a generator
that can be used with \code{\link[=get_frame]{get_frame()}} and \code{\link[=get_raw_frames]{get_raw_frames()}} to extract frames
for a specific time point scaled between 0 and 1.
}
\examples{
d <- data.frame(
  x = runif(20),
  y = runif(20),
  time = runif(20),
  duration = runif(20, max = 0.1)
)
from_left <- function(x) {
  x$x <- -0.5
  x
}
to_right <- function(x) {
  x$x <- 1.5
  x
}

gen <- gen_events(d, 'cubic-in-out', start = time, end = time + duration,
                  enter = from_left, exit = to_right, enter_length = 0.1,
                  exit_length = 0.05)

get_frame(gen, 0.65)

}
\seealso{
Other Other generators: 
\code{\link{gen_along}()},
\code{\link{gen_at}()},
\code{\link{gen_components}()},
\code{\link{gen_keyframe}()}
}
\concept{Other generators}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_state.R
\name{.get_last_frame}
\alias{.get_last_frame}
\alias{.get_first_frame}
\alias{.with_prior_frames}
\alias{.with_later_frames}
\alias{.has_frames}
\title{Helpers for working with tweened data}
\usage{
.get_last_frame(data)

.get_first_frame(data)

.with_prior_frames(prior, new_tween, nframes)

.with_later_frames(later, new_tween, nframes)

.has_frames(data)
}
\arguments{
\item{data, prior, later}{A data.frame. If a \code{.frame} column exists it will be interpreted
as a data.frame containing multiple states}

\item{new_tween}{The result of a tweening}
}
\value{
A data.frame
}
\description{
These are internal helpers for extracting and inserting data into a
data.frame of tweened states.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_elements.R
\name{tween_elements}
\alias{tween_elements}
\title{Create frames based on individual element states}
\usage{
tween_elements(data, time, group, ease, timerange, nframes)
}
\arguments{
\item{data}{A data.frame consisting at least of a column giving the
observation id, a column giving timepoints for each state and a column giving
the easing to apply when transitioning away from the state.}

\item{time}{The name of the column holding timepoints}

\item{group}{The name of the column holding the observation id}

\item{ease}{The name of the column holding the easing function name}

\item{timerange}{The range of time to span. If missing it will default to
\code{range(data[[time]])}}

\item{nframes}{The number of frames to generate. If missing it will default
to \code{ceiling(diff(timerange) + 1)} (At least one frame for each
individual timepoint)}
}
\value{
A data.frame with the same columns as \code{data} except for the
group and ease columns, but replicated \code{nframes} times. Two additional
columns called \code{.frame} and \code{.group} will be added giving the frame
number and observation id for each row.
}
\description{
This function creates tweens for each observation individually, in cases
where the data doesn't pass through collective states but consists of fully
independent transitions. Each observation is identified by an id and each
state must have a time associated with it.
}
\examples{
data <- data.frame(
  x = c(1, 2, 2, 1, 2, 2),
  y = c(1, 2, 2, 2, 1, 1),
  time = c(1, 4, 10, 4, 8, 10),
  group = c(1, 1, 1, 2, 2, 2),
  ease = rep('cubic-in-out', 6)
)

data <- tween_elements(data, 'time', 'group', 'ease', nframes = 100)

}
\seealso{
Other data.frame tween: 
\code{\link{tween_along}()},
\code{\link{tween_appear}()},
\code{\link{tween_components}()},
\code{\link{tween_events}()},
\code{\link{tween_states}()}
}
\concept{data.frame tween}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display_ease.R
\name{display_ease}
\alias{display_ease}
\title{Display an easing function}
\usage{
display_ease(ease)
}
\arguments{
\item{ease}{The name of the easing function to display (see details)}
}
\value{
This function is called for its side effects
}
\description{
This simple helper lets you explore how the different easing functions govern
the interpolation of data.
}
\details{
How transitions proceed between states are defined by an easing function. The
easing function converts the parameterized progression from one state to the
next to a new number between 0 and 1. \code{linear} easing is equivalent to
an identity function that returns the input unchanged. In addition there are
a range of additional easers available, each with three modifiers.

\strong{Easing modifiers:}
\describe{
\item{-in}{The easing function is applied as-is}
\item{-out}{The easing function is applied in reverse}
\item{-in-out}{The first half of the transition it is applied as-is, while
in the last half it is reversed}
}

\strong{Easing functions}
\describe{
\item{quadratic}{Models a power-of-2 function}
\item{cubic}{Models a power-of-3 function}
\item{quartic}{Models a power-of-4 function}
\item{quintic}{Models a power-of-5 function}
\item{sine}{Models a sine function}
\item{circular}{Models a pi/2 circle arc}
\item{exponential}{Models an exponential function}
\item{elastic}{Models an elastic release of energy}
\item{back}{Models a pullback and relase}
\item{bounce}{Models the bouncing of a ball}
}

In addition to this function a good animated explanation can be found
\href{https://easings.net}{here}.
}
\examples{
# The default - identity
display_ease('linear')

# A more fancy easer
display_ease('elastic-in')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gen_keyframe.R
\name{interpolate_custom_at}
\alias{interpolate_custom_at}
\title{Fallback for keyframe vector support}
\usage{
interpolate_custom_at(from, to, at, ease)
}
\arguments{
\item{from, to}{vectors to interpolate between}

\item{at}{value between 0 and 1 defining the point}

\item{ease}{the easing function to use}
}
\description{
Fallback for keyframe vector support
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_state.R
\name{tween_state}
\alias{tween_state}
\alias{keep_state}
\alias{open_state}
\alias{close_state}
\title{Compose tweening between states}
\usage{
tween_state(.data, to, ease, nframes, id = NULL, enter = NULL, exit = NULL)

keep_state(.data, nframes)

open_state(.data, ease, nframes, enter)

close_state(.data, ease, nframes, exit)
}
\arguments{
\item{.data}{A data.frame to start from. If \code{.data} is the result of a prior
tween, only the last frame will be used for the tween. The new tween will
then be added to the prior tween}

\item{to}{A data.frame to end at. It must contain the same columns as .data
(exluding \code{.frame})}

\item{ease}{The easing function to use. Either a single string or one for
each column in the data set.}

\item{nframes}{The number of frames to calculate for the tween}

\item{id}{The column to match observations on. If \code{NULL} observations will be
matched by position. See the \emph{Match, Enter, and Exit} section for more
information.}

\item{enter, exit}{functions that calculate a start state for new observations
that appear in \code{to} or an end state for observations that are not present in
\code{to}. If \code{NULL} the new/old observations will not be part of the tween. The
function gets a data.frame with either the start state of the exiting
observations, or the end state of the entering observations and must return
a modified version of that data.frame. See the \emph{Match, Enter, and Exit}
section for more information.}
}
\value{
A data.frame containing all the intermediary states in the tween,
each state will be enumerated by the \code{.frame} column
}
\description{
The \code{tween_state()} is a counterpart to \code{tween_states()} that is aimed at
letting you gradually build up a scene by composing state changes one by one.
This setup lets you take more control over each state change and allows you
to work with datasets with uneven number of rows, flexibly specifying what
should happen with entering and exiting data. \code{keep_state()} is a simple
helper for letting you pause at a state. \code{open_state()} is a shortcut from
tweening from an empty dataset with a given \code{enter()} function while
\code{close_state()} is the same but will instead tween into an empty dataset with
a given \code{exit()} function.
}
\section{Match, Enter, and Exit}{

When there are discrepancies between the two states to tweeen between you
need a way to resolve the discrepancy before calculating the intermediary
states. With discrepancies we mean that some data points are present in the
start state and not in the end state, and/or some are present in the end
state but not in the start state. A simple example is that the start state
contains 100 rows and the end state contains 70. There are 30 missing rows
that we need to do something about before we can calculate the tween.

\strong{Making pairs}
The first question to answer is "How do we know which observations are
disappearing (\emph{exiting}) and/or appearing (\emph{entering})?". This is done with
the \code{id} argument which should give a column name to match rows between the
two states on. If \code{id = NULL} the rows will be matched by position (in the
above example the last 30 rows in the start state will be entering). The \code{id}
column must only contain unique values in order to work.

\strong{Making up states}
Once the rows in each state has been paired you'll end up with three sets of
data. One containing rows that is present in both the start and end state,
one containing rows only present in the start state, and one only containing
rows present in the end state. The first group is easy - here you just tween
between each rows - but for the other two we'll need some state to start or
end the tween with. This is really the purpose of the \code{enter} and \code{exit}
functions. They take a data frame containing the subset of data that has not
been matched and must return a new data frame giving the state that these
rows must be tweened from/into. A simple example could be an \code{enter} function
that sets the variable giving the opacity in the plot to 0 - this will make
the new points fade into view during the transition.

\strong{Ignoring discrepancies}
The default values for \code{enter} and \code{exit} is \code{NULL}. This value indicate that
non-matching rows should simply be ignored for the transition and simply
appear in the last frame of the tween. This is the default.
}

\examples{
data1 <- data.frame(
  x = 1:20,
  y = 0,
  colour = 'forestgreen',
  stringsAsFactors = FALSE
)
data2 <- data1
data2$x <- 20:1
data2$y <- 1

data <- data1 \%>\%
  tween_state(data2, 'linear', 50) \%>\%
  keep_state(20) \%>\%
  tween_state(data1, 'bounce-out', 50)

# Using enter and exit (made up numbers)
df1 <- data.frame(
  country = c('Denmark', 'Sweden', 'Norway'),
  population = c(5e6, 10e6, 3.5e6)
)
df2 <- data.frame(
  country = c('Denmark', 'Sweden', 'Norway', 'Finland'),
  population = c(6e6, 10.5e6, 4e6, 3e6)
)
df3 <- data.frame(
  country = c('Denmark', 'Norway'),
  population = c(10e6, 6e6)
)
to_zero <- function(x) {
  x$population <- 0
  x
}
pop_devel <- df1 \%>\%
  tween_state(df2, 'cubic-in-out', 50, id = country, enter = to_zero) \%>\%
  tween_state(df3, 'cubic-in-out', 50, id = country, enter = to_zero,
              exit = to_zero)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_states.R
\name{tween_states}
\alias{tween_states}
\title{Tween a list of data.frames representing states}
\usage{
tween_states(data, tweenlength, statelength, ease, nframes)
}
\arguments{
\item{data}{A list of data.frames. Each data.frame must contain the same
number of rows, but only the first data.frame needs to contain all columns.
Subsequent data.frames need only contain the columns that shows change.}

\item{tweenlength}{The lengths of the transitions between each state.}

\item{statelength}{The length of the pause at each state.}

\item{ease}{The easing functions to use for the transitions. See details.}

\item{nframes}{The number of frames to generate. The actual number of frames
might end up being higher depending on the regularity of \code{tweenlength}
and \code{statelength}.}
}
\value{
A data.frame with the same columns as the first data.frame in
\code{data}, but replicated \code{nframes} times. An additional column called
\code{.frame} will be added giving the frame number.
}
\description{
This function is intended to create smooth transitions between states of
data. States are defined as full data.frames or data.frames containing only
the columns with change. Each state can have a defined period of pause, the
transition length between each states can be defined as well as the easing
function.
}
\examples{
data1 <- data.frame(
  x = 1:20,
  y = 0,
  colour = 'forestgreen',
  stringsAsFactors = FALSE
)
data2 <- data1
data2$x <- 20:1
data2$y <- 1

data <- tween_states(list(data1, data2), 3, 1, 'cubic-in-out', 100)

}
\seealso{
Other data.frame tween: 
\code{\link{tween_along}()},
\code{\link{tween_appear}()},
\code{\link{tween_components}()},
\code{\link{tween_elements}()},
\code{\link{tween_events}()}
}
\concept{data.frame tween}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R, R/gen.R, R/gen_keyframe.R
\docType{data}
\name{col_classes}
\alias{col_classes}
\alias{gen_internal}
\alias{PHASE_LEVELS}
\alias{is_generator}
\alias{generator_settings}
\alias{generator_settings<-}
\alias{ease_type}
\alias{ease_type<-}
\alias{col_types}
\alias{col_types<-}
\alias{gen_data}
\alias{gen_data<-}
\alias{gen_to_data_frame}
\alias{data_frame_to_gen}
\alias{keyframes}
\alias{keyframes<-}
\alias{frame_times}
\alias{frame_times<-}
\title{Generator internals}
\format{
An object of class \code{character} of length 5.
}
\usage{
col_classes(data)

PHASE_LEVELS

is_generator(x)

generator_settings(x)

generator_settings(x) <- value

ease_type(x)

ease_type(x) <- value

col_types(x)

col_types(x) <- value

gen_data(x)

gen_data(x) <- value

gen_to_data_frame(...)

data_frame_to_gen(x)

keyframes(x)

keyframes(x) <- value

frame_times(x)

frame_times(x) <- value
}
\arguments{
\item{x}{A generator object}
}
\value{
Various data
}
\description{
Generator internals
}
\keyword{datasets}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa.R
\name{vec_tween_class}
\alias{vec_tween_class}
\title{Get the nominal class of a vector}
\usage{
vec_tween_class(x)
}
\arguments{
\item{x}{a vector}
}
\description{
Get the nominal class of a vector
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_fill.R
\name{tween_fill}
\alias{tween_fill}
\title{Fill out missing values by interpolation}
\usage{
tween_fill(data, ease)
}
\arguments{
\item{data}{A data.frame or vector.}

\item{ease}{A character vector giving valid easing functions. Recycled to
match the ncol of \code{data}}
}
\value{
If \code{data} is a data.frame then a data.frame with the same
columns. If \code{data} is a vector then a vector.
}
\description{
This tween fills out \code{NA} elements (or \code{NULL} elements if \code{data} is a list)
by interpolating between the prior and next non-missing values.
}
\examples{
# Single vector
tween_fill(c(1, NA, NA, NA, NA, NA, 2, 6, NA, NA, NA, -2), 'cubic-in-out')

# Data frame
tween_fill(mtcars[c(1, NA, NA, NA, NA, 4, NA, NA, NA, 10), ], 'cubic-in')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tweenr_package.R
\docType{package}
\name{tweenr-package}
\alias{tweenr}
\alias{tweenr-package}
\title{tweenr: Interpolate Data for Smooth Animations}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

In order to create smooth animation between states of data, tweening is necessary. This package provides a range of functions for creating tweened data that can be used as basis for animation. Furthermore it adds a number of vectorized interpolaters for common R data types such as numeric, date and colour.
}
\details{
tweenr is a small collection of functions to help you in creating
intermediary representations of your data, i.e. interpolating states of data.
As such it's a great match for packages such as animate and gganimate, since
it can work directly with data.frames of data, but it also provide fast and
efficient interpolaters for numeric, date, datetime and colour that are
vectorized and thus more efficient to use than the build in interpolation
functions (mainly \code{\link[stats:approxfun]{stats::approx()}} and
\code{\link[grDevices:colorRamp]{grDevices::colorRamp()}}).

The main functions for data.frames are \code{\link[=tween_states]{tween_states()}},
\code{\link[=tween_elements]{tween_elements()}} and \code{\link[=tween_appear]{tween_appear()}}, while the
standard interpolaters can be found at \code{\link[=tween]{tween()}}
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/thomasp85/tweenr}
  \item Report bugs at \url{https://github.com/thomasp85/tweenr/issues}
}

}
\author{
\strong{Maintainer}: Thomas Lin Pedersen \email{thomasp85@gmail.com} (\href{https://orcid.org/0000-0002-5147-4711}{ORCID})

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tween_at.R
\name{tween_at}
\alias{tween_at}
\title{Get a specific position between two states}
\usage{
tween_at(from, to, at, ease)
}
\arguments{
\item{from, to}{A data.frame or vector of the same type. If either is of
length/nrow 1 it will get repeated to match the length of the other}

\item{at}{A numeric between 0 and 1 recycled to match the nrow/length of
\code{from}}

\item{ease}{A character vector giving valid easing functions. Recycled to
match the ncol of \code{from}}
}
\value{
If \code{from}/\code{to} is a data.frame then a data.frame with the same
columns. If \code{from}/\code{to} is a vector then a vector.
}
\description{
This tween allows you to query a specific postion between two states rather
than generate evenly spaced states. It can work with either data.frames or
single vectors and each row/element can have its own position and easing.
}
\examples{
tween_at(mtcars[1:6, ], mtcars[6:1, ], runif(6), 'cubic-in-out')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gen_keyframe.R
\name{gen_keyframe}
\alias{gen_keyframe}
\alias{add_pause}
\alias{add_keyframe}
\title{Generator for keyframe based tweening}
\usage{
gen_keyframe(keyframe = NULL, pause = 0)

add_pause(.data, pause = 0)

add_keyframe(
  .data,
  keyframe,
  ease,
  length,
  id = NULL,
  enter = NULL,
  exit = NULL
)
}
\arguments{
\item{keyframe}{A data frame to use as a keyframe state}

\item{pause}{The length of the pause at the current keyframe}

\item{.data}{A data.frame to start from. If \code{.data} is the result of a prior
tween, only the last frame will be used for the tween. The new tween will
then be added to the prior tween}

\item{ease}{The easing function to use. Either a single string or one for
each column in the data set.}

\item{length}{The length of the transition}

\item{id}{The column to match observations on. If \code{NULL} observations will be
matched by position. See the \emph{Match, Enter, and Exit} section for more
information.}

\item{enter, exit}{functions that calculate a start state for new observations
that appear in \code{to} or an end state for observations that are not present in
\code{to}. If \code{NULL} the new/old observations will not be part of the tween. The
function gets a data.frame with either the start state of the exiting
observations, or the end state of the entering observations and must return
a modified version of that data.frame. See the \emph{Match, Enter, and Exit}
section for more information.}
}
\value{
A \code{keyframe_generator} object
}
\description{
This is a generator version of \code{\link[=tween_state]{tween_state()}} and its utility functions. It
returns a generator that can be used with \code{\link[=get_frame]{get_frame()}} and
\code{\link[=get_raw_frames]{get_raw_frames()}} to extract frames for a specific time point scaled between
0 and 1.
}
\examples{
df1 <- data.frame(
  country = c('Denmark', 'Sweden', 'Norway'),
  population = c(5e6, 10e6, 3.5e6)
)
df2 <- data.frame(
  country = c('Denmark', 'Sweden', 'Norway', 'Finland'),
  population = c(6e6, 10.5e6, 4e6, 3e6)
)
df3 <- data.frame(
  country = c('Denmark', 'Norway'),
  population = c(10e6, 6e6)
)
to_zero <- function(x) {
  x$population <- 0
  x
}
gen <- gen_keyframe(df1, 10) \%>\%
  add_keyframe(df2, 'cubic-in-out', 35, id = country, enter = to_zero) \%>\%
  add_pause(10) \%>\%
  add_keyframe(df3, 'cubic-in-out', 35, id = country, enter = to_zero,
               exit = to_zero) \%>\%
  add_pause(10)

get_frame(gen, 0.25)
}
\seealso{
Other Other generators: 
\code{\link{gen_along}()},
\code{\link{gen_at}()},
\code{\link{gen_components}()},
\code{\link{gen_events}()}
}
\concept{Other generators}
